"""
Copyright (C) 2015-2018  Axel Rau <axel.rau@chaos1.de>

This file is part of serverPKI.

serverPKI is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Foobar is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with serverPKI.  If not, see <http://www.gnu.org/licenses/>.
"""

# commandline interface module

import time
import sys

from script import path

from DSKM.utils import options as opts
import DSKM.config as conf
            
import DSKM.logger as logger
import DSKM.registrar as reg
import DSKM.zone as zone
import DSKM.misc as misc

def execute_from_command_line():

    
    print('operate_dskm started.')
    ##print('External scondaries are {}.'.format(conf.external_secondaries))
    print('Options are {}.'.format(opts))
    
    
    root = path(conf.ROOT_PATH)
    if opts.debug: opts.verbose = True
    l = logger.Logger(opts.verbose, opts.debug, opts.cron)
    
    
    if opts.registrar_status or opts.query_status or \
            opts.purge_all_registrar_completion_info:
        cl = reg.getResultList(opts.query_status)
        if not cl:
            l.logError('Failed')
            return 1
        if opts.purge_all_registrar_completion_info:
            for line in cl['result']:
                if len(line) < 10: continue
                tid = line.split(' ')[1]
                if reg.deleteResult(tid):
                    print('.', end='', flush=true)
            print()
        elif opts.registrar_status:
            print('- timestamp -- ---------- Tracking-Id --------- Proc-ID  --- task ---  domain    result')
            for line in cl['result']:
                print(line)
        else:
            for k in sorted(cl.keys()):
                if k != 'result':
                    print('%s:\t%s' % (k, cl[k]))
        return 0
    
    if not root.exists:
        l.logWarn('No key root directory; creating one.')
        root.mkdir(mode=0o750)
        root = path(conf.ROOT_PATH)
    if opts.verbose:
        print('')
    l.logVerbose('Scanning ' + root)
    
    root.cd()
    root = path('.')
    for dir in root.list('*'):
        if dir.is_dir:
            misc.zone_dirs.append(dir.name)
    misc.zone_dirs.sort(key = len)
    misc.zone_dirs.reverse()
    
    if opts.stopSigningOfZone:
        zone_name = opts.stopSigningOfZone
        print('[Stopping signing of %s]' % zone_name)
        if zone_name in misc.zone_dirs:
            try:
                z = zone.managedZone(zone_name)
                res1 = z.stopSigning(opts.force)
                print('[Set dnssec-secure-to-insecure to yes in zone config of named.conf]')
                print('[Do "cd <zone_dir>; rm *.jbk *.jnl *.signed ; sleep 1 ; rndc stop ; rndc start"]')
                print('[...repeat until no DNSKEYs and RRSIGs remain in zone]')
                return res1
            except misc.AbortedZone:
                print('?Failed to stop signing of zone ' + zone_name)
                return 1
            except misc.CompletedZone:
                print('%%Unsigned zone ' + zone_name)
                return 0
            pass
        else:
            print('?%s not a managed zone.' % opts.stopSigningOfZone)
        return 1
    
    l.logDebug('[ Doing zones: ]')
    l.logDebug( misc.zone_dirs )

    if opts.test_registrar_DS_submission:
        for zone_name in misc.zone_dirs:
            try:
                misc.zones[zone_name] = zone.managedZone(zone_name)
            except misc.AbortedZone as a:
                print(a.data)
                print('%Skipping zone ' + zone_name)
            except misc.CompletedZone:
                pass
            
            try:
                reg.regTest(misc.zones[zone_name], opts.dry_run)
            except misc.AbortedZone as a:
                print(a.data)
                print('%Skipping zone ' + zone_name)
            except misc.CompletedZone:
                pass
        return 0
    for zone_name in misc.zone_dirs:
        try:
            misc.zones[zone_name] = zone.managedZone(zone_name)
            if not misc.zones[zone_name].verifySerial(): continue
            misc.zones[zone_name].performStateTransition()
            misc.zones[zone_name].validate()
        except misc.AbortedZone as a:
            print(a.data)
            print('%Skipping zone ' + zone_name)
        except misc.CompletedZone:
            pass
    l.mailErrors()
    


"""
 DSKM DNSsec Key Management
 
 Copyright (c) 2012 Axel Rau, axel.rau@chaos1.de

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

# -----------------------------------------
REG/ripe.py - Interface module to the Regional Internet Registry RIPE's REST API 
          see https://labs.ripe.net/ripe-database/database-api/api-documentation
"""
# -----------------------------------------
from datetime import date, datetime
import http.client
import pprint
import ssl
import time
import urllib.parse
import xml.etree.ElementTree as etree

import sys
# -----------------------------------------

# -----------------------------------------
# Globals
# -----------------------------------------
theConnection = None
theSecureConnection = None

# -----------------------------------------

import DSKM.logger as logger
l = logger.Logger()
# -----------------------------------------


# -----------------------------------------
# Configurables
# -----------------------------------------
import DSKM.config as conf
#------------------------------------------------------------------------------

#--------------------------
#   classes
#--------------------------
class ConnectionRipe():
    """Connection to Ripe.net whois-db server"""
    
    _singleton = None
    conn = 0
    
    
    def __new__(cls, *args, **kwargs):
        if not cls._singleton:
            cls._singleton = super(ConnectionRipe, cls ).__new__(cls, *args, **kwargs)
        return cls._singleton
    
    
    def __init__(self):
        
        ConnectionRipe.conn = http.client.HTTPConnection(conf.registrar['Ripe']['server'])
        ##ConnectionRipe.conn.set_debuglevel(1)
        l.logDebug('Connected.')
        
    def conn(self):
        return ConnectionRipe.conn

class SecureConnectionRipe():
    """Connection to Ripe.net whois-db server"""
    
    _singleton = None
    conn = 0
    
    
    def __new__(cls, *args, **kwargs):
        if not cls._singleton:
            cls._singleton = super(SecureConnectionRipe, cls ).__new__(cls, *args, **kwargs)
        return cls._singleton
    
    
    def __init__(self):
        
        context = ssl.SSLContext(ssl.PROTOCOL_TLSv1)
        SecureConnectionRipe.conn = http.client.HTTPSConnection(conf.registrar['Ripe']['server'], context = context)
        ##SecureConnectionRipe.conn.set_debuglevel(10)
        l.logDebug('Securely connected.')
        
    def conn(self):
        return SecureConnectionRipe.conn

# -----------------------------------------
# Functions
# -----------------------------------------
def regRemoveAllDS(zone_name):

    return makeRequest(zone_name, ())

def regAddDS(zone_name, args, test_run, dry_run):

    return makeRequest(zone_name, args, test_run, dry_run)


def getResultList(rid):
    # no result list with Ripe
    return ''

# -----------------------------------------
# Internal functions
# -----------------------------------------

def dbQuery(url):
    
    c = SecureConnectionRipe().conn
    
    u = '/' + url
    l.logDebug('dbQuery: Host: %s GET %s' % (conf.registrar['Ripe']['server'], u))
    c.request('GET', u)
    
    with c.getresponse() as r1:
        t = etree.fromstring(r1.read())
        if r1.status != 200:
            l.logError('Query of RIPE.NET whois DB server failed , because: % - HTTP - status: %' %
                (r1.reason, repr(r1.status)))
    if l.debug:
        print('[Dumping XML tree of dbQuery response:]')
        etree.dump(t)
    return t

def dbRequest(method, url, dry_run, body=None, headers={}):
    
    c = SecureConnectionRipe().conn
    
    dr = ''
    if dry_run: dr = '&dry-run'
    u = '/' + url + '?password=' + conf.registrar['Ripe']['account_pw'] + dr
    # l.logDebug seems dysfunctional here:
    l.logDebug('dbRequest(method,url,,,): %s %s' % (method, u))
    l.logDebug('dbRequest(,,,headers,): %s' % (repr(headers)))
    l.logDebug('dbRequest(,,body,,): %s' % (body))
    l.logDebug('dbRequest(method,url,,,): %s %s' % (method, u))
    c.request(method, u, body, headers)
    
    with c.getresponse() as r1:
        t = etree.fromstring(r1.read())
        if r1.status != 200:
            l.logError('Update of RIPE.NET whois DB server failed , because: %s - HTTP - status: %s' %
                (r1.reason, repr(r1.status)))
    if l.debug:
        print('[Dumping XML tree of dbRequest response:]')
        etree.dump(t)
    return t

def makeRequest(zone_name, args, test_run, dry_run):   # construct and perform the request for Ripe
    newChangedValue = ''
    ns = set()

    received_tree = None
    domain_atts = {}
        
    if zone_name.split('.')[-1] not in 'arpa':
        l.logError('Internal inconsitency: Zone not supported by Ripe: "%s"' % (zone_name))
        return None

    newChangedValue = changedTimestamp()    # sets up value of 'changed' attribute
    try:
        # query current config
        received_tree = dbQuery('search?source=ripe&query-string=' + zone_name + '&flags=no-filtering')
        assert received_tree != None
    except:
        l.logError('Request query DS-RR of zone %s at Ripe failed' % (zone_name))
        return None
    
    domain_atts = extract_domain_atts(received_tree)
    if not extract_and_report_error_messages(received_tree):
        return None
    
    for arg in args:
        if (None, '') in (arg['tag'], arg['alg'], arg['digest_type'], str(arg['digest'])):
            l.logError('Internal inconsistency: regAddDS(): at least one argument of key %d is empty: "%d","%d","%s"'
                % (arg['tag'], arg['alg'], arg['digest_type'], str(arg['digest'])))
            return None
        ns.add(str('%d %d %d %s' % (arg['tag'], arg['alg'], arg['digest_type'], str(arg['digest']))))
        
    if test_run and len(ns) > 0:
        domain_atts['ds-rdata'] = list(ns)
    else:
        (changed_something, domain_atts) = updateDomainAtts(domain_atts, ns, newChangedValue)
        if not changed_something:
            l.logWarn('No changes needed at registrar Ripe')
            return {'TID': ''}
    
    h = {}
    h['Content-Type'] = 'application/xml'
    b = etree.tostring(create_new_tree(domain_atts), encoding="unicode")
    try:
        # request update of config
        received_tree = dbRequest('PUT', 'ripe/domain/' + zone_name, dry_run, body=b, headers=h)
        assert received_tree != None
    except:
        l.logError('Request update DS-RR of zone %s to Ripe failed' % (zone_name))
        return None
    domain_atts = extract_domain_atts(received_tree)
    if not extract_and_report_error_messages(received_tree):
        return None
    ts = set(domain_atts['ds-rdata'])
    if ns != ts:
        l.logError('DS-RR of zone %s returned from Ripe differ from request' % (zone_name))
        return None
    return {'TID': newChangedValue}
    
def extract_domain_atts(received_tree):
    domain_atts = {}
    for obj in received_tree.findall('objects/*'):
        for (k,v) in obj.attrib.items():
            if k == 'type' and v == 'domain':
                for att in obj.findall('attributes/*'):
                    name = att.get('name')
                    value = att.get('value')
                    ## RIPE-NCC introduced last-modified on 2015-06, which must not be provided
                    if name == 'last-modified':
                        continue
                    print('%s: %s' % (name, value))
                    if name in domain_atts:
                        domain_atts[name].append(value)
                    else:
                        domain_atts[name] = [value]
                        
    for k in domain_atts:
        domain_atts[k].sort()
    return domain_atts

def extract_and_report_error_messages(received_tree):
    error_messages = []
    no_error = True
    for em in received_tree.findall('errormessages/*'):
        if em.tag == 'errormessage':
            err = False
            severity = em.get('severity')
            if severity == 'Error':
                no_error = False
                err = True
            arg_list = []
            arg_list.append(em.get('text'))
            args = em.findall('args')
            if args:
                for arg in args:
                    a = arg.get('value')
                    arg_list.append(a)
            for arg in arg_list:
                if err:
                    l.logError(arg)
                else:
                    l.logWarn(arg)
    return no_error

def updateDomainAtts(domain_atts, ns, newChangedValue):
    os = set(domain_atts['ds-rdata']) if 'ds-rdata' in domain_atts else set()
    if os == ns:
        return (False, domain_atts)
    else:
        domain_atts['ds-rdata'] = list(ns)
        ## RIPE-NCC removed changed on 2015-05
        ##if newChangedValue not in domain_atts['changed']:
        ##    domain_atts['changed'].append(newChangedValue)
        return (True, domain_atts)

def create_new_tree(domain_atts):
    root = etree.Element('whois-resources')
    ose = etree.SubElement(root, 'objects')
    oe = etree.SubElement(ose, 'object')
    oe.set('type', 'domain')
    se = etree.SubElement(oe, 'source')
    se.set('id', 'ripe')
    ase = etree.SubElement(oe, 'attributes')
    
    de = etree.Element('attribute')     #will become domain attribute
    for k in domain_atts:
        for v in domain_atts[k]:
            if k == 'domain':
                de.set('name', k)
                de.set('value', v)
            else:
                a = etree.SubElement(ase, 'attribute')
                a.set('name', k)
                a.set('value', v)
    ase.insert(0, de)                   #domain attribute must be first
    ##etree.dump(root)
    return root

def changedTimestamp():
    newChangedValue = ''
    timestamp = datetime.now()
    newChangedValue = str('%s %s' % (conf.registrar['Ripe']['changed_email'], timestamp.strftime('%Y%m%d')))
    return newChangedValue




"""
 DSKM DNSsec Key Management
 
 Copyright (c) 2012 Axel Rau, axel.rau@chaos1.de

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

# -----------------------------------------
key.py - SigningKey class module

# -----------------------------------------
KSK roll over best practice from ttp://tools.ietf.org/html/draft-ietf-dnsop-rfc4641bis-10

4.1.2.  Key Signing Key Rollovers

   For the rollover of a Key Signing Key, the same considerations as for
   the rollover of a Zone Signing Key apply.  However, we can use a
   Double Signature scheme to guarantee that old data (only the apex key
   set) in caches can be verified with a new key set and vice versa.
   Since only the key set is signed with a KSK, zone size considerations
   do not apply.

Kolkman & Mekking        Expires October 1, 2012               [Page 22]
 
Internet-Draft   DNSSEC Operational Practices, Version 2      March 2012


   ---------------------------------------------------------------------
    initial            new DNSKEY        DS change    DNSKEY removal
   ---------------------------------------------------------------------
   Parent:
    SOA_0 -----------------------------> SOA_1 ------------------------>
    RRSIG_par(SOA) --------------------> RRSIG_par(SOA) --------------->
    DS_K_1 ----------------------------> DS_K_2 ----------------------->
    RRSIG_par(DS) ---------------------> RRSIG_par(DS) ---------------->

   Child:
    SOA_0              SOA_1 -----------------------> SOA_2
    RRSIG_Z_10(SOA)    RRSIG_Z_10(SOA) -------------> RRSIG_Z_10(SOA)

    DNSKEY_K_1         DNSKEY_K_1 ------------------>
                       DNSKEY_K_2 ------------------> DNSKEY_K_2
    DNSKEY_Z_10        DNSKEY_Z_10 -----------------> DNSKEY_Z_10
    RRSIG_K_1(DNSKEY)  RRSIG_K_1 (DNSKEY) ---------->
                       RRSIG_K_2 (DNSKEY) ----------> RRSIG_K_2(DNSKEY)
   ---------------------------------------------------------------------

    Figure 4: Stages of Deployment for a Double Signature Key  Signing
                               Key Rollover

   initial:  Initial version of the zone.  The parental DS points to
      DNSKEY_K_1.  Before the rollover starts, the child will have to
      verify what the TTL is of the DS RR that points to DNSKEY_K_1 --
      it is needed during the rollover and we refer to the value as
      TTL_DS.

   new DNSKEY:  During the "new DNSKEY" phase, the zone administrator
      generates a second KSK, DNSKEY_K_2.  The key is provided to the
      parent, and the child will have to wait until a new DS RR has been
      generated that points to DNSKEY_K_2.  After that DS RR has been
      published on all servers authoritative for the parent's zone, the
      zone administrator has to wait at least TTL_DS to make sure that
      the old DS RR has expired from caches.

   DS change:  The parent replaces DS_K_1 with DS_K_2.

   DNSKEY removal:  DNSKEY_K_1 has been removed.

   The scenario above puts the responsibility for maintaining a valid
   chain of trust with the child.  It also is based on the premise that
   the parent only has one DS RR (per algorithm) per zone.  An
   alternative mechanism has been considered.  Using an established
   trust relation, the interaction can be performed in-band, and the
   removal of the keys by the child can possibly be signaled by the
   parent.  In this mechanism, there are periods where there are two DS
   RRs at the parent.

"""

import sys

from script import path, shell, opts
import script
import fnmatch
from datetime import date, datetime
import time

# for salt
from Crypto import Random as rand
import binascii

import copy

import dns.resolver, dns.message, dns.query, dns.rdatatype, dns.rdtypes.ANY.DNSKEY, dns.rcode
import dns.dnssec, dns.zone

import json
import os
from shutil import chown

import re

import pprint

# -----------------------------------------

import DSKM.logger as logger
l = logger.Logger()
import DSKM.misc as misc

# -----------------------------------------


# -----------------------------------------
# Configurables
# -----------------------------------------
import DSKM.config as conf

if conf.ZSK_A_I_INTERVAL <= conf.ZSK_P_A_INTERVAL + \
    conf.ZSK_I1_A2_INTERVAL + conf.ZSK_I_D_INTERVAL:
    l.logError('Configuration error: ZSK_A_I_INTERVAL must be greater than ZSK_P_A_INTERVAL + ZSK_I1_A2_INTERVAL + ZSK_I_D_INTERVAL')
    sys.exit(1)

if conf.KSK_A_I_INTERVAL <= conf.KSK_P_A_INTERVAL + \
    conf.KSK_I1_A2_INTERVAL + conf.KSK_I_D_INTERVAL:
    l.logError('Configuration error: KSK_A_I_INTERVAL must be greater than KSK_P_A_INTERVAL + KSK_I1_A2_INTERVAL + KSK_I_D_INTERVAL')
    sys.exit(1)

#------------------------------------------------------------------------------

master_resolver = dns.resolver.Resolver()
master_resolver.lifetime = conf.NS_TIMEOUT
master_resolver.nameservers = conf.master
master_resolver.use_edns(edns=0, ednsflags=0, payload=4096)

secondary_resolver = dns.resolver.Resolver()
secondary_resolver.lifetime = conf.NS_TIMEOUT
secondary_resolver.nameservers = conf.external_secondaries
secondary_resolver.use_edns(edns=0, ednsflags=0, payload=4096)



#--------------------------
#   classes
#--------------------------


class SigningKey(object):
    """SigningKey"""

    global KSTT, ZSTT, l
    l = logger.Logger()                  # the logger
    
    def __init__(self, task, name, file_name, sender, nsec3=False, cloneFromKeyInactiveAt=0):
        
        self.name = name
        self.file_name = None
        self.zone = sender
        self.nsec3 = nsec3

        self.type = None
        
        self.algo = conf.KEY_ALGO_NSEC
        if nsec3: self.algo = conf.KEY_ALGO_NSEC3
        
        # values read from key file
        self.timingData = {}
        self.keytag = ''        # key tag
        self.dnssec_flags = 0   # flags
        self.sepkey = 0         # sep flag =KSK)
        self.dnssec_alg = 0     # key algorithm
        self.pubkey_base64 = ''

        self.dsHash = [None,None]  # 2 DS hashes
        
        self.mypath = path(conf.ROOT_PATH + '/' + name)
        self.mypath.cd()
        l.logDebug('Instantiating SigningKey; pwd=%s.' % (self.mypath))

        #-----------------------------
        # functions in SigningKey.__init__
        #-----------------------------
        # Read key meta data from key file
        def readKey(keyFileName):
            
            #   Read timing meta data from key
            def readKeyTimingData(keyFileName, type):
                result = None
                if not type in 'PAID':
                    l.logError('Internal inconsistency: readKeyTimingData called with wrong type ' 
                        + type + ' for key ' +keyFileName)
                    e = misc.AbortedZone("")
                    raise e
                try:
                    (rubbish, result) = str(shell(conf.BIND_TOOLS + 'dnssec-settime   -u -p ' + type + ' ' + keyFileName, stdout='PIPE').stdout).split(None)
                except script.CommandFailed:
                    l.logError('Error from dnssec_settime while reading timing data of '  +keyFileName)
                    e = misc.AbortedZone("")
                    raise e
                if result == 'UNSET':
                    return 0
                else:
                    return int(result)
            
            fd = None
            try:
                fd = open(keyFileName, 'r')
            except IOError:
                (exc_type, exc_value, exc_traceback) = sys.exc_info()
                errmsg = "?Can't open key file, because %s" % (exc_value)
                print(errmsg)
                e = misc.AbortedZone(errmsg)
                raise e
            flags = None
            l.logDebug('readKey(%s) opened file.' % (keyFileName))
            for line in fd:
                ##l.logDebug('Keyfile line is %s' % (line, ))
                (name, ttl, dns_class, rr, flags, x) = line.split(None, 5)
                if name == ';':
                    continue
                ##l.logDebug('Keyfile line found with name=%s ttl=%s class=%s RR=%s flags=%s\n%s' % (name, ttl, dns_class, rr, flags, x))
                if dns_class == 'IN' and rr == 'DNSKEY':
                    if __debug__:
                        if not self.name == name[:-1]: raise AssertionError('?Wrong domain name in keyfile %s (%s/%s)' % (keyFileName, name, self.name))
                    st = line.find('DNSKEY ')
                    tok = dns.tokenizer.Tokenizer(line[st+7:])
                    dnskey_rdata = dns.rdtypes.ANY.DNSKEY.DNSKEY.from_text(dns.rdataclass.ANY, dns.rdatatype.DNSKEY, tok, origin=name, relativize=False)
                    self.keytag = dns.dnssec.key_id(dnskey_rdata)
                    self.dnssec_flags = dnskey_rdata.flags
                    self.sepkey = self.dnssec_flags & 0x1;
                    self.dnssec_alg = dnskey_rdata.algorithm
                    ##self.pubkey_base64 = dns.rdata._base64ify(dnskey_rdata.key)
                    self.pubkey_base64 = dns.rdata._base64ify(dnskey_rdata.key, chunksize=2000)
                    
                    l.logDebug('Read DNSSEC key id=%d with flags=%d alg=%d' % (self.keytag, self.dnssec_flags, self.dnssec_alg))
                else:
                    l.logError('Unrecognized line "%s" in key file: ' % (line, keyFileName))
                    e = misc.AbortedZone("")
                    raise e
                if flags == '257':
                    l.logDebug('Key ' + keyFileName + ' is KSK')
                    if self.sepkey == 0:
                        l.logError('Inkonsistent sep flag found in %s' % (keyFileName))
                        e = misc.AbortedZone("")
                        raise e
                    self.type = 'KSK'
                    break
                elif flags == '256':
                    l.logDebug('Key ' + keyFileName + ' is ZSK')
                    if self.sepkey == 1:
                        l.logError('Inkonsistent sep flag found in %s' % (keyFileName))
                        e = misc.AbortedZone("")
                        raise e
                    self.type = 'ZSK'
                    break
                else:
                    l.logError('Key neither KSK not ZSK: ' + keyFileName)
                    e = misc.AbortedZone("")
                    raise e
            
            fd.close()
            
            self.timingData['P'] = readKeyTimingData(keyFileName, 'P')
            self.timingData['A'] = readKeyTimingData(keyFileName, 'A')
            self.timingData['I'] = readKeyTimingData(keyFileName, 'I')
            self.timingData['D'] = readKeyTimingData(keyFileName, 'D')
            
            if self.type == 'KSK' and self.zone.pcfg['Registrar'] != 'Local':
                self.digestOfDS()

        #-----------------------------
        # end of functions in SigningKey.__init__
        #-----------------------------

        l.logDebug('Creating SigningKey instance task=%s, name=%s, file_name=%s, nsec3=%s, cloneFromKeyInactiveAt=%d'
                                                            % (task, name, file_name, nsec3, cloneFromKeyInactiveAt))
        if task == 'read':
            self.file_name = file_name
            readKey(file_name)
        elif task == 'ZSK':             # active when predecessor inactive
            inactive_from_now = self.zone.pcfg['Timing']['zsk']['ai']               # 1st key is immediately active
            delete_from_now = inactive_from_now + self.zone.pcfg['Timing']['zsk']['id']
            s = conf.BIND_TOOLS + 'dnssec-keygen -a ' + self.algo + ' -b ' + repr(conf.KEY_SIZE_ZSK) + ' -n ZONE ' \
                + '-A +' + str(0) + 'd ' +'-I +' + repr(inactive_from_now) + 'd ' \
                + '-D +' + repr(delete_from_now) +'d -L ' + repr(conf.TTL_DNSKEY) + ' ' + name
            sn = s                      # preserve in case creation of cloned key fails
            if cloneFromKeyInactiveAt != 0:
                prepublishInterval = cloneFromKeyInactiveAt - int(time.time()) - 60 # active - now (seconds)
                if prepublishInterval <= 0:      # did we wait too long? (new active > old inactive ?)
                    prepublishInterval = int(time.time()) - 60                      # yes: publish in 1 minute
                    l.logWarn('Failed timely key rollover of %s before %s' % (file_name, datetime.fromtimestamp(cloneFromKeyInactiveAt).isoformat()))
                prepublishIntervalDays = int(1 + prepublishInterval/(3600*24))
                inactive_from_now = prepublishIntervalDays + self.zone.pcfg['Timing']['zsk']['ai']  # prep. + inact. - act.
                delete_from_now = inactive_from_now + self.zone.pcfg['Timing']['zsk']['id']
                s = conf.BIND_TOOLS + 'dnssec-keygen -S ' + file_name + ' -i +' + repr(prepublishInterval) + ' -I +' \
                + repr(inactive_from_now) + 'd ' \
                    + '-D +' + repr(delete_from_now) +'d -L ' + repr(conf.TTL_DNSKEY)
            l.logDebug(s)
            try:
                result = shell(s, stdout='PIPE').stdout.strip()
            except script.CommandFailed:
                l.logWarn('Error while creating cloned ZSK for ' + name)
                l.logWarn('Trashed Timing data in parent ?')
                if s != sn:
                    l.logWarn('Retrying uncloned create of ZSK')
                    l.logDebug(sn)
                    try:
                        result = shell(sn, stdout='PIPE').stdout.strip()
                    except script.CommandFailed:
                        l.logError('Error while creating ZSK for ' + name)
                        e = misc.AbortedZone("")
                        raise e
                else:
                    e = misc.AbortedZone("")
                    raise e
            self.file_name = result + '.key'
            if conf.OWNER_OF_PRIVATE_KEY != '':
                chown(result + '.key',conf.OWNER_OF_PRIVATE_KEY)
                chown(result + '.private',conf.OWNER_OF_PRIVATE_KEY)
            print('[Key ' + self.file_name + ' created.]')
            readKey(self.file_name)
        elif task == 'KSK':         # active now
            active_from_now = 0
            if cloneFromKeyInactiveAt != 0: active_from_now = self.zone.pcfg['Timing']['ksk']['pa']
            
            inactive_from_now = active_from_now + self.zone.pcfg['Timing']['ksk']['ai']
            delete_from_now = inactive_from_now + self.zone.pcfg['Timing']['ksk']['id']
            s = conf.BIND_TOOLS + 'dnssec-keygen -a ' + self.algo + ' -b ' + repr(conf.KEY_SIZE_KSK) + ' -n ZONE -f KSK ' \
                + '-P +0d -A +' + repr(active_from_now) + 'd -I +' + repr(inactive_from_now) + 'd ' \
                + '-D +' + repr(delete_from_now) + 'd -L ' + repr(conf.TTL_DNSKEY) + ' ' + name
            if cloneFromKeyInactiveAt != 0 and cloneFromKeyInactiveAt < int(time.time()):
                l.logWarn('Failed timely key rollover of %s before %s' % (file_name, datetime.fromtimestamp(cloneFromKeyInactiveAt).isoformat()))
            l.logDebug(s)
            try:
                result = shell(s, stdout='PIPE').stdout.strip()
            except script.CommandFailed:
                l.logError('Error while creating KSK for ' + name)
                e = misc.AbortedZone("")
                raise e
            self.file_name = result + '.key'
            if conf.OWNER_OF_PRIVATE_KEY != '':
                chown(result + '.key',conf.OWNER_OF_PRIVATE_KEY)
                chown(result + '.private',conf.OWNER_OF_PRIVATE_KEY)
            print('[Key ' + self.file_name + ' created.]')
            readKey(self.file_name)
        else:
            l.logError('Internal inconsitency: SigningKey instantiating  with wrong task ' + task)
            e = misc.AbortedZone("")
            raise e
        l.logVerbose('%s' % self.__str__())
        l.logDebug('Instantiated public key %s' % self.pubkey_base64)

    def __str__(self):
        def getKeyTimingData(type):
            if self.timingData[type] == 0:
                return 'UNSET'
            else:
                return datetime.fromtimestamp(self.timingData[type]).isoformat()
        
        return str('%s/%s/%d/%d(A:%s, I:%s, D:%s)' % (self.name, self.type, self.keytag, self.zone.pstat[self.type.lower()]['State'],
                    getKeyTimingData('A'), getKeyTimingData('I'), getKeyTimingData('D')))
    
    def digestOfDS(self):
        self.mypath.cd()
        def read1DS(i):
            digest = ''
            s = str(conf.BIND_TOOLS + 'dnssec-dsfromkey -%d %s' % (i, self.file_name))
            try:                                      
                result = shell(s, stdout='PIPE').stdout.strip()
                l.logDebug('digestOfDS(): dnssec-dsfromkey returned: "%s"' % (result,))
                rr = result.split(None)
                if len(rr) == 7:
                    digest = rr[6]
                elif len(rr) == 8:
                    digest = rr[6]+rr[7]
                else:
                    assert('?Wrong result from dnssec-dsfromkey: %s' % (rr,))
                l.logDebug('digestOfDS(): returning: "%s"' % digest)
                return digest                
            except script.CommandFailed:              
                l.logError('Error while creating DS RR for ' + self.name)
                e = misc.AbortedZone("")
                raise e
        
        if self.type != 'KSK':
            l.logError("Can't create/delete DS from ZSK (internal inconsitency)" + self.name)
            e = misc.AbortedZone("")
            raise e
        if not self.dsHash[0]:
            self.dsHash[0] = read1DS(1)
            self.dsHash[1] = read1DS(2)
        return self.dsHash
    
    def UpdateDS(self, activity, secondKey):    # create/submit/retire signer RR from KSK
        l.logDebug('UpdateDS(%s, %s) called' % (activity, repr(secondKey)))
        if self.type != 'KSK':
            l.logError("Can't create/delete DS from ZSK (internal inconsitency)" + self.name)
            e = misc.AbortedZone("")
            raise e
        if self.zone.pcfg['Registrar'] == 'Local':
            if self.zone.parent_dir != None:
                return self.UpdateLocalDS(activity, secondKey)
            return True                         # nothing to do with local domain w/o parent
        else:
            l.logDebug('Calling UpdateRemoteDS(%s) called' % (activity))
            return self.zone.UpdateRemoteDS(activity, self.keytag)
    
    def UpdateLocalDS(self, activity, secondKey):    # create/submit/retire signer RR from KSK
        self.mypath.cd()
        result = ''
        
        if activity != 'retire' and activity != 'delete':
            l.logVerbose('Creating DS-RR from KSK %s' % self.__str__())
            s = conf.BIND_TOOLS + 'dnssec-dsfromkey ' + conf.DIGEST_ALGO_DS + ' ' + self.file_name
            l.logDebug(s)                   
            try:                                      
                result = shell(s, stdout='PIPE').stdout.strip()
                result = result + '\n'
            except script.CommandFailed:              
                l.logError('Error while creating DS RR for ' + self.name)
                e = misc.AbortedZone("")
                raise e
        elif activity == 'retire':
            l.logVerbose('Deleting local DS-RR from KSK %s' % self.__str__())
        else:
            l.logVerbose('Deleting all local DS-RRs from zone %s' % self.name)
        
        ds_file_name = self.zone.parent_dir + '/' + self.name + '.ds'
        l.logDebug('Updating local DS-RR stored in "%s"' % (ds_file_name,))
        old = ''
        try:
            if secondKey:
                with open(ds_file_name, 'r', encoding="ASCII") as fd:
                    if activity == 'retire':
                        lines = fd.readlines()
                        n = 1
                        if conf.DIGEST_ALGO_DS == '':   # 2 DS-RR (SHA1 + SHA256)?
                            n = 2
                        old = ''.join(lines[n:])
                    elif activity == 'delete':
                        pass
                    else:
                        old = fd.read()
            with open(ds_file_name, 'w', encoding="ASCII") as fd:
                fd.write(old + result)
        except:
            (exc_type, exc_value, exc_traceback) = sys.exc_info()
            l.logError('Error while reading/writing local DS-RR file for KSK %s \n\tbecause %s' % (self.__str__(), exc_value))
            e = misc.AbortedZone("")
            raise e
        self.updateSOA(self.zone.parent_dir + '/' + self.zone.parent + '.zone')
        return True        
    
    def updateSOA(self, filename): # update serial of SOA in zone file
        timestamp = datetime.now()
        current_date = timestamp.strftime('%Y%m%d')
        self.mypath.cd()
        
        zf = ''
        with open(filename, 'r', encoding="ASCII") as fd:
            try:
                zf = fd.read()
            except:                 # file not found or not readable
                l.logError("Can't read zone file " + filename)
                e = misc.AbortedZone('')
                raise e
        ##l.logDebug('Updating SOA: zone file before update:' + zf)
        sea = re.search('(\d{8})(\d{2})(\s*;\s*)(Serial number)', zf)
        old_date = sea.group(1)
        daily_change = sea.group(2)
        if old_date == current_date:
           daily_change = str('%02d' % (int(daily_change) +1, ))
        else:
            daily_change = '01'
        zf = re.sub('\d{10}', current_date + daily_change, zf, count=1)
        ##l.logDebug('Updating SOA: zone file after update:' + zf)
        with open(filename, 'w', encoding="ASCII") as fd:
            try:
                fd.write(zf)
            except:                 # file not found or not readable
                l.logError("Can't write zone file " + filename)
                e = misc.AbortedZone('')
                raise e
        try:
             res = str(shell('rndc reload', stderr='PIPE').stderr)
             l.logDebug('Rndc reload returned: %s' % (res))
        except script.CommandFailed:
             l.logError('Error while reloading zones after updating SOA of %s ( %s )' % (self.name, res))
             e = misc.AbortedZone("")
             raise e
                                                    # check if condition for state transition is true and do action if so
    def state_transition(self, secondKey):          # secondKey is true if we are 2nd KSK/ZSK
        
        l.logDebug('state_transition(%s) called for %s %s' % (secondKey, self.keytag, self.type))
        state = -1                                  # state is index into state table
        stt = None                                  # state table
        key = ''                                    # key for accessing state
        if self.type == 'KSK':                      # key signing key
            stt = KSTT                              # use corresponding state table
            key = 'ksk'                             # and state key
        else:                                       # zone signing key
            stt = ZSTT
            key = 'zsk'
        state = self.zone.pstat[key]['State']
        l.logDebug('State is %d (%s)' % (self.zone.pstat[key]['State'],stt[state]['s']))
        if state == -1:                             # initial state: begin signing
            state = 0
            self.zone.pstat[key]['State'] = state   # we have an SigningKey instance: advance state
            l.logDebug('New state is %d (%s)' % (self.zone.pstat[key]['State'],stt[state]['s']))
            l.logVerbose('State transition of %s/%s from -1 to %d(%s) after %s retries' %
                (self.name, self.type, self.zone.pstat[key]['State'], stt[state]['s'], self.zone.pstat[key]['Retries']))
            return True
                                                    # not initial state
        if stt[state]['c'](self, stt[state]['ca'], secondKey):      # check condition for state transition
            if 'a' in stt[state].keys():                            # succeeded: action present?
                if 'aa' in stt[state].keys():                       # yes, call it - arg present?
                    stt[state]['a'](self, stt[state]['aa'], secondKey)  # yes, call it with arg
                else:
                    stt[state]['a'](self, secondKey)
            if 'ns' in stt[state].keys():           # action routine returned True or none called: 'next state' key present?
                state =  stt[state]['ns']           # yes, use it
            else:
                state = state + 1                   # no, increment it
            l.logVerbose('State transition of %s/%d from %d to %d (%s) after %s retries' %
                (self.type, self.keytag,
                self.zone.pstat[key]['State'], state, stt[state]['s'], self.zone.pstat[key]['Retries']))
            self.zone.pstat[key]['State'] = state
            self.zone.pstat[key]['Retries'] = 0
            return True                             # we had a transition
        if not secondKey and l.cronjob():
            self.zone.pstat[key]['Retries'] = str(int(self.zone.pstat[key]['Retries']) + 1)
        timeout = stt[state]['t']
        if timeout == 'N':      # no timeout
            return False
        key_type = stt[state]['ca']
        if not key_type:
            return False
        if '1' in key_type and secondKey or '2' in key_type and not secondKey:
            return False

        if timeout == 'S':    # short timeout
            if int(self.zone.pstat[key]['Retries']) > int(conf.TIMEOUT_SHORT):
                l.logWarn('Timeout [S] of state transition for %s/%s/%d at state %d (%s) after %s retries' %
                ((self.name, self.type, self.keytag,
                state, stt[state]['s'], self.zone.pstat[key]['Retries'])))
            return False
        elif 'I' in timeout or 'D' in timeout:
            if not self.fixedTimeout(timeout):
                return False
            l.logWarn('Timeout [%s] of state transition for %s/%s/%d at state %d (%s) after %s retries' %
            ((timeout, self.name, self.type, self.keytag,
            state, stt[state]['s'], self.zone.pstat[key]['Retries'])))
            return False
        elif timeout == 'P':    # long timeout
            if int(self.zone.pstat[key]['Retries']) > \
                int(conf.TIMEOUT_PREPUB_ADDITION) + int(self.zone.pcfg['Timing']['ksk']['id']) * int(conf.CRON_FREQ):
                l.logWarn('Timeout [P] of state transition for %s/%s/%d at state %d (%s) after %s retries' %
                ((self.name, self.type, self.keytag,
                state, stt[state]['s'], self.zone.pstat[key]['Retries'])))
            return False
        l.logWarn('Wrong statetable timeout value "%s"' % (timeout))
        return False                                # we stay in current state
    
    def activeTime(self):
        return self.timingData['A']
    
    def inactiveTime(self):
        return self.timingData['I']
    
    def deleteTime(self):
        return self.timingData['D']
    
    def fixedTimeout(self, time_type):                   # returns True, if fixed timeout has happened
        l.logDebug('fixedTimeout(' + time_type + ') called')
        myTime = self.timingData[time_type] + conf.TIMEOUT_SHORT * 3600  # timeout time
        now = int(time.time())
        l.logDebug('fixedTimeout: myTime=%s; now=%s' % (
                        datetime.fromtimestamp(myTime).isoformat(), datetime.fromtimestamp(now).isoformat()))
        if myTime <= now:
            return True
        return False                                    
    
    # -----------------------------
    # Actions on state transitions in SigningKey (immediately after one test below returns true)
    # -----------------------------
    def create_a(self, key_type, secondKey):     # create follow up KSK or ZSK
        l.logDebug('create_a(key_type)')
        return self.zone.createFollowUpKey(self)
    
    def delete_a(self, key_type, secondKey):
        l.logDebug('delete_a(key_type) called')
        my_tag = self.keytag
        if key_type == 'delete_all':
            my_tag = 0
        return self.zone.markForDeletion(my_tag)
    
    def rename_a(self, key_type, secondKey):
        l.logDebug('rename_a(key_type) called')
        return True   # nothing to do (keyfile 1 deleted previously and next round only key file 2 present)
    
    def submit_ds(self, activity, secondKey):
        l.logDebug('submit_ds(%s) called' % (activity))
        return self.UpdateDS(activity, secondKey)
    
    def set_delete_time(self, secondKey):
        l.logDebug('set_delete_time() called')
        self.mypath.cd()                # change to zone directory
        try:
             (rubbish, result) = str(shell(conf.BIND_TOOLS + 'dnssec-settime   -D +' + str(self.zone.pcfg['Timing']['ksk']['id']) + 'd ' + self.file_name, stdout='PIPE').stdout).split(None)
        except script.CommandFailed:
             l.logError('Error from dnssec_settime while setting delete time of '  +keyFileName)
             e = misc.AbortedZone("")
             raise e
        return True
    
    # -----------------------------
    # Tests for state transitions in SigningKey
    # -----------------------------
    def test_if_included(self, key_type, secondKey):    # test, if included in zone by our master
        global master_resolver,secondary_resolver       # included means: used for signing
        
        l.logDebug('test_if_included(' + key_type + ', ' + str(secondKey) + ') called with name %s' % (self.name))
        if '1' in key_type and secondKey or '2' in key_type and not secondKey:
            return False
       
        l.logDebug('test_if_included(' + key_type + ', ' + str(secondKey) + ') testing %s' % (self.name))
        
        r = master_resolver
        if 'ds' in key_type:
            if self.zone.pcfg['Registrar'] != 'Local':  # DS maintained by registrar?
                r = misc.authResolver(self.zone.parent) # yes - use resolver bound to their auth NS
            elif not self.zone.parent_dir:              # locally maintained - do we have a parent?
                return True                             # no - no parent: no DS - state test always succeeds
            l.logDebug('test_if_included(): List of auth NS to query: %s' % (repr(r.nameservers)))
            try:
                res = r.query(self.name, 'DS')
            except dns.resolver.NoAnswer:
                return False
            except (dns.exception.Timeout, dns.resolver.NXDOMAIN):
                (exc_type, exc_value, exc_traceback) = sys.exc_info()
                errmsg = "%s: DS query timed out or gave NXDOMAIN. %s, %s" % \
                    (self.name, exc_type, exc_value)
                l.logError(errmsg)
                e = misc.AbortedZone('? ' + errmsg)
                raise e
            else:
                for ds in res.rrset.items:
                    keytag = ds.key_tag
                    l.logDebug('test_if_included("ds", secondKey) matching keytag: %s == %s' % (keytag, self.keytag))
                    if keytag == self.keytag:
                        l.logDebug('test_if_included("ds", secondKey) RRSIG matched ourselves')
                        return True
            return False
        else:                                           # testing for signed DNSKEY or SOA 
            l.logDebug('test_if_included(): 0')
            try:
                """
                s = None
                if self.zone.pcfg['Registrar'] == 'Local':  # zone maintained local?
                    s = conf.master[0]
                else:
                    s = conf.external_secondaries[-2]       # use 2nd to last of our secondaries for now **FIXME**
                zone = dns.zone.from_xfr(dns.query.xfr(s, self.name, relativize=False, lifetime=30.0), relativize=False)
                my_covers = dns.rdatatype.DNSKEY        # DNSKEYs signed by KSK
                if self.type == 'ZSK':
                   my_covers = dns.rdatatype.SOA        # others signed by ZSK
                ##import pdb;pdb.set_trace()
                rds = zone.find_rrset(self.name + '.', 'RRSIG', covers=my_covers)
                print(repr(rds))
                if False:                               # need to learn how redirect pprint to string
                    dbgmsg1 = 'zone.find.rrset:\n'
                    dbgmsg2 = ''
                    pp = pprint.PrettyPrinter(indent=4,stream=dbgmsg2)
                    pp.pprint(rds)
                    l.logDebug(dbgmsg1 + dbgmsg2)
                
                for rrsig_rdata in rds.items:
                    key_tag = rrsig_rdata.key_tag
                    l.logDebug('test_if_included(key_type, secondKey) matching keytag: %s == %s' % (key_tag, self.keytag))
                    if key_tag == self.keytag:
                        l.logDebug('test_if_included(key_type, secondKey) RRSIG matched ourselves')
                        return True                 # at least one RR signed by ourselves
                """
                ns = None
                if self.zone.pcfg['Registrar'] == 'Local':  # zone maintained local?
                    ns = conf.master[0]
                else:
                    ns = conf.external_secondaries[-2]      # use 2nd to last of our secondaries for now **FIXME**
                my_covers = dns.rdatatype.DNSKEY       # DNSKEYs signed by KSK
                if self.type == 'ZSK':
                    my_covers = dns.rdatatype.SOA      # others signed by ZSK
                q = dns.message.make_query (self.name, my_covers, want_dnssec=True)
                l.logDebug('test_if_included(): Querying %s from %s' % (dns.rdatatype._by_value[my_covers], ns))
                my_answer = dns.query.tcp (q, ns, conf.NS_TIMEOUT)
                for rdata in my_answer.answer:
                    for item in rdata.items:
                        if item.rdtype == dns.rdatatype.RRSIG:
                            if item.covers() == my_covers:
                                if my_covers == dns.rdatatype.SOA:
                                    l.logDebug('test_if_included() RRSIG matched ourselves')
                                    return True
                                else:
                                    l.logDebug('test_if_included(matching keytag: %s == %s' % (item.key_tag, self.keytag))
                                    if item.key_tag == self.keytag:
                                        return True
                
            except (dns.resolver.NoAnswer, KeyError): # KeyError if no RRSIG of type <covers> and class IN exist
                (exc_type, exc_value, exc_traceback) = sys.exc_info()
                errmsg = str("%s: RRSIG query returned NoAnswer or KeyError\n(%s/%s/%s)"
                    % (self.name, exc_type, exc_value, exc_traceback))
                l.logDebug(errmsg)
            except (dns.resolver.NXDOMAIN):         # should never occur
                errmsg = "%s: RRSIG query returned domain %s none-existent" % (self.name,)
                l.logError(errmsg)
                e = misc.AbortedZone('? ' + errmsg)
                raise e
            except (dns.exception.Timeout, IOError):
                (exc_type, exc_value, exc_traceback) = sys.exc_info()
                errmsg = "%s: RRSIG query network error: (Timeout or connection refused). %s, %s" % \
                    (self.name, exc_type, exc_value)
                l.logError(errmsg)
                e = misc.AbortedZone('? ' + errmsg)
                raise e
            except (dns.exception.FormError):
                (exc_type, exc_value, exc_traceback) = sys.exc_info()
                errmsg = "%s: RRSIG query Format error: (FormError). %s, %s" % \
                    (self.name, exc_type, exc_value)
                l.logError(errmsg)
                e = misc.AbortedZone('? ' + errmsg)
                raise e
        return False
    
    def test_if_excluded(self, key_type, secondKey):       # test, if excluded from zone by our master
        l.logDebug('test_if_excluded(' + key_type + ') called') # (no longer used for signing)
        return not self.test_if_included(key_type, secondKey)
    
    def test_if_deleted(self, key_type, secondKey):         # test if DNSKEY has been deleted from RRset by master
        l.logDebug('test_if_deleted: keytype: {} secondKey: {}'.format(key_type, secondKey))
        if '1' in key_type and secondKey or '2' in key_type and not secondKey:
            l.logDebug('test_if_deleted: NoOP')
            return False
       
        l.logDebug('test_if_deleted(' + key_type + ') called with name %s' % (self.name))
        
        keytags = self.masters_DNSKEYs()
        if self.keytag in keytags:
            l.logDebug('test_if_deleted(' + key_type + ') returning False with name %s' % (self.name))
            return False
        l.logDebug('test_if_deleted(' + key_type + ') returning True with name %s' % (self.name))
        return True
        
    def masters_DNSKEYs(self):  # prevent from bruteforcing named by querying too often for DNSKEY
        if len(self.zone.master_DNSKEY_cache) == 0:
            r = master_resolver
            res = None
            try:
                res = r.query(self.name, 'DNSKEY')
            except dns.resolver.NoAnswer:
                l.logError('masters_DNSKEYs got NOANSWER while querying for DNSKEY of %s' % (self.name))
                sys.exit(1)
            except dns.resolver.NXDOMAIN:
                l.logError('masters_DNSKEYs got NXDOMAIN while querying for DNSKEY of %s' % (self.name))
                sys.exit(1)
            except (dns.exception.Timeout):
                l.logError('masters_DNSKEYs got timeout while querying for DNSKEY of %s' % (self.name))
                sys.exit(1)
            else:                                           # no exception
                for dnskey_rdata in res.rrset.items:
                    keytag = dns.dnssec.key_id(dnskey_rdata)
                    self.zone.master_DNSKEY_cache.append(keytag)
        return self.zone.master_DNSKEY_cache
        
    
    def test_if_time_reached(self, time_type, secondKey):  # test, if arbitrary point in time reached
        
        if '1' in time_type and secondKey or '2' in time_type and not secondKey:
            return False
        
        myTime = 0
        l.logDebug('test_if_time_reached(' + time_type + ') called')
        if time_type == 'zsk1_followup':
            myTime = self.timingData['I'] - self.zone.pcfg['Timing']['zsk']['i1a2'] * 3600 * 24 - \
            self.zone.pcfg['Timing']['zsk']['pa'] * 3600 * 24               # rollover + prepublish time before ZSK1 inactive
        elif time_type == 'zsk1_inactive':
            myTime = self.timingData['I']                                           # ZSK1 inactive = ZSK2 active
        elif time_type == 'zsk1_delete':
            myTime = self.timingData['D'] + conf.TIMEOUT_SHORT * 3600               # ZSK1 delete time reached
        
        elif time_type == 'ds1_submit':                         # DS to be submitted prepublish interval after active
            myTime = self.timingData['A'] + self.zone.pcfg['Timing']['ksk']['pa'] * 3600 * 24 # pa is pre publish interval
        elif time_type == 'ksk1_followup':
            myTime = self.timingData['I'] - self.zone.pcfg['Timing']['ksk']['i1a2'] * 3600 * 24 - \
                    self.zone.pcfg['Timing']['ksk']['pa'] * 3600 * 24 # rollover + prepublish time before KSK1 inactive
        elif time_type == 'ds2_submit':                               # DS to be submitted prepublish interval after active
            myTime = self.timingData['A'] + self.zone.pcfg['Timing']['ksk']['pa'] * 3600 * 24 # pa is pre publish interval
        elif time_type == 'ksk1_inactive':
            myTime = self.timingData['I']                                           # KSK1 inactive
        elif time_type == 'ksk1_delete':
            myTime = self.timingData['D'] + conf.TIMEOUT_SHORT * 3600               # KSK1 to be deleted
        elif time_type == 'ksk_delete':
            myTime = self.timingData['D'] + self.zone.pcfg['Timing']['ksk']['pa'] * 3600 * 24 # prepub after DS retirement
        else:
            if __debug__:
                raise AssertionError('?Internal error: test_if_time_reached called with wrong argument "%s"' % (time_type,))
            return False
        
        now = int(time.time())
        l.logDebug('test_if_time_reached: myTime=%s; now=%s' % (
                        datetime.fromtimestamp(myTime).isoformat(), datetime.fromtimestamp(now).isoformat()))
        if myTime <= now:
            return True
        return False 
    
    # -----------------------------
    # State tables in SigningKey
    # -----------------------------
    #          s = state        t = timeout c = check cond. for transition, ca = argument for c, a = action aa=arg  ns = next state
    ZSTT = (
            { 's': 'ZSK1 created', 't':'S','c': test_if_included,      'ca': 'zsk1',                                        },
            { 's': 'ZSK1 active',  't':'N','c': test_if_time_reached,  'ca': 'zsk1_followup','a': create_a,  'aa': 'zsk'    },
            { 's': 'ZSK2 created', 't':'P','c': test_if_included,      'ca': 'zsk2',                                        },
            { 's': 'ZSK2 active',  't':'S','c': test_if_excluded,      'ca': 'zsk1',                                        },
            { 's': 'ZSK1 inactive','t':'D','c': test_if_deleted,       'ca': 'zsk1',         'a': delete_a,  'aa': 'zsk',  'ns': 1,}
    )
    zsk_state_max = 4
    
    KSTT = (
            { 's': 'KSK1 created', 't':'S','c': test_if_included,      'ca': 'ksk1',                                        }, # 0
            { 's': 'KSK1 active',  't':'N','c': test_if_time_reached,  'ca': 'ds1_submit',  'a': submit_ds, 'aa': 'publish1'}, # 1
            { 's': 'DS1 submitted','t':'S','c': test_if_included,      'ca': 'ds1',                                         }, # 2
            { 's': 'DS1 published','t':'N','c': test_if_time_reached,  'ca': 'ksk1_followup','a': create_a,  'aa': 'ksk'    }, # 3
            { 's': 'KSK2 created', 't':'P','c': test_if_included,      'ca': 'ksk2',        'a': submit_ds, 'aa': 'publish2'}, # 4
            { 's': 'KSK2 active',  't':'S','c': test_if_included,      'ca': 'ds2',         'a': submit_ds, 'aa': 'retire'  }, # 5
            { 's': 'DS2 published','t':'S','c': test_if_excluded,      'ca': 'ds1',                                         }, # 6
            { 's': 'DS1 retired',  't':'I','c': test_if_excluded,      'ca': 'ksk1'                                         }, # 7
            { 's': 'KSK1 inactive','t':'D','c': test_if_deleted,       'ca': 'ksk1',     'a': delete_a,  'aa': 'ksk', 'ns':3}, # 8
            { 's': 'DS retire request submitted',\
                                   't':'S','c':test_if_excluded,'ca':'ds',         'a': set_delete_time,           }, # 9
            { 's': 'DS retired',   't':'N','c': test_if_time_reached,  'ca': 'ksk_delete',  'a': delete_a,  'aa': 'delete_all','ns':-1}
    )
    ksk_state_max = 8

    
    #-----------------------------
    # functions in SigningKey
    #-----------------------------
    def createNSEC3PARAM():
        """
        http://strotmann.de/roller/dnsworkshop/entry/take_your_dnssec_with_a
        """
        salt = binascii.b2a_hex(rand.get_random_bytes(6)).decode('ASCII').upper()



"""
 DSKM DNSsec Key Management
 
 Copyright (c) 2012 Axel Rau, axel.rau@chaos1.de

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

# -----------------------------------------
REG/joker.py - Interface module to the registrar Joker.com Domain Management API 
          see https://dmapi.joker.com
"""


# -----------------------------------------
import http.client
import ssl
import time
import urllib.parse

import sys
# -----------------------------------------

# -----------------------------------------
# Globals
# -----------------------------------------
theConnection = None
# -----------------------------------------

import DSKM.logger as logger
l = logger.Logger()
# -----------------------------------------


# -----------------------------------------
# Configurables
# -----------------------------------------
import DSKM.config as conf
#------------------------------------------------------------------------------

#--------------------------
#   classes
#--------------------------
class ConnectionJoker(object):
    """Connection to Joker.com DMAPI server"""
    global l
    def __init__(self):
        self.myConnection = 0
        self.session = {}
        self.sslContext = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH, cafile=conf.ca_file)
        try:
            self.myConnection = http.client.HTTPSConnection(conf.registrar['Joker']['server'], context=self.sslContext)
            l.logDebug('HTTPS- connection succeeded. Authentication follows.')
            self.myConnection.request("GET", '/request/login?username=' + 
                conf.registrar['Joker']['account_name'] + '&password=' + conf.registrar['Joker']['account_pw'])
            l.logDebug('Authentication initiated.')
        except Exception:
            l.logError('Failed to connect to Joker.com DMAPI server, because ', str(sys.exc_info()[1]))
            return None
        
        r1 = self.myConnection.getresponse()
        l.logDebug('Got authentication response.')
        if r1.status != 200:
            l.logError('Failed to connect to Joker.com DMAPI server, because: ' + r1.reason)
        while not r1.closed:
            for line in r1.read().decode('ASCII').splitlines():
                l.logDebug('Got line from Joker: {}'.format(line))
                k = v = ''
                if len(line) > 2:
                    try:
                        (kv) =  line.split(':')
                        k = kv[0]
                        v = ''.join(kv[1:])
                    except (ValueError):
                        l.logError('Parsing error while reading data from Joker.com. Line: {}'.format(line))
                        return None
                    self.session[k] = v.strip()
                    l.logDebug('Response: ' + k + ': ' + v)
                else:
                    l.logDebug('Got line shorter then 2 chars: ' + line)
                    break
            l.logDebug('No more lines to read from response.')
            break
        l.logDebug('Collected authentication response.')
        if self.session['Status-Code'] != '0':
            l.logError('Failed to sign-on at Joker.com, because: ' + self.session['Status-Text'])
            for k in session.keys():
                print(k, ': ',  self.session[k])
            return None
        l.logDebug('Authentication succeeded.')
    
    def conn(self):
        return self.myConnection
    
    def sid(self):
        return self.session['Auth-Sid']


# -----------------------------------------
# Functions
# -----------------------------------------
def requestJoker(query_string):
    
    status = {}
    result = []
    header_done = False
    global theConnection
    
    if not theConnection:
        theConnection = ConnectionJoker()
        if not theConnection: return None
    c = theConnection.conn()
    if not c:  return None
    if '?' in query_string:
        s = '/request/' + query_string + '&auth-sid=' + theConnection.sid()
    else:
        s = '/request/' + query_string + '?auth-sid=' + theConnection.sid()
    l.logDebug(s)
    try:
        c.request("GET", s)
    except Exception:
        l.logError('Request to Joker.com DMAPI server failed, because: ', str(sys.exc_info()[1]))
        l.logError('Query was: "' + query_string + '"')
        return None
    r1 = c.getresponse()
    l.logDebug('Got response.')
    if r1.status != 200:
        l.logError('Request to Joker.com DMAPI server failed , because: ' + r1.reason + ' - HTTP - status ' + repr(r1.status))
    while not r1.closed:
        for line in r1.read().decode('ASCII').splitlines():
            items = line.split(':')
            if len(items) == 2:
                (k,v) = line.split(':')
                status[k] = v.strip()
                l.logDebug('Response: ' + k + ': ' + v)
            else:
                result.append(line)
                l.logDebug('Line: ' + line)

        break
    if status['Status-Code'] != '0':
        lines = []
        for k in status.keys():
            lines.append(str('%s:\t%s\n' % (k, status[k])))
        l.logDebug('Request to Joker.com DMAPI server failed, because:\n%s' % ''.join(lines))
        l.logError('Request to Joker.com DMAPI server failed')
        return None
    status['result'] = result        
    return status

def regRemoveAllDS(zone):
    if zone.split('.')[-1] not in 'arpa':
        try:
            q = 'domain-modify?domain=' + zone + '&dnssec=0'
            cl = requestJoker(q)
            if not cl: return None
            for c in cl:
                l.logDebug(c + ':   ' + str(cl[c]))
            if 'Tracking-Id' in cl:
                stat = getResult(cl['Tracking-Id'])
                if 'request_state' in stat and stat['request_state'] == 'SUCCESS':
                    return stat
                else:
                    for c in stat:
                        l.logDebug(c + ':   ' + str(stat[c]))
                    l.logError('Unsuccesfull request state received.')
                    return None
            else:
                l.logError('Missing Tracking-Id in response from Joker while removing DS-RR of %s' % zone)
                return None
            l.logError('Request remove all DS-RR of zone %s to Joker failed, because: %s' % (zone, stat['result_msg']))
            l.logError('Order Number is %s, and request, sent to Joker was "%s"' % (stat['order number'], q))
            printStatus(stat, ('ds-', 'result_'))
            return None
        except (KeyError):
            l.logError('Request remove all DS-RR of zone %s to Joker failed' % (zone))
            return None
    l.logError('Internal inconsitency: regRemoveAllDS(): DS-Removal of .arpa. not implemented')
    return None
    
def regAddDS(zone, args):
    stat = None
    tld = zone.split('.')[-1]
    if tld not in 'arpa':
        try:
            if tld != 'de':
                q = str('domain-modify?domain=%s&dnssec=1' % zone)
                i = 1
                for arg in args:
                    if (None, '') in (arg['tag'], arg['alg'], arg['digest_type'], str(arg['digest'])):
                        l.logError('Internal inconsitency: regAddDS(): at least one argument of key %d is empty: "%d","%d","%s"'
                            % (arg['tag'], arg['alg'], arg['digest_type'], urllib.parse.quote(str(arg['digest']))))
                        return None
                    q = q + str('&ds-%d=%d:%d:%d:%s' %
                        (i, arg['tag'], arg['alg'], arg['digest_type'], urllib.parse.quote(str(arg['digest']))))
                    i = i + 1
                cl = requestJoker(q)
                if not cl: return None
                for c in cl:
                    l.logDebug(c + ':   ' + str(cl[c]))
                if 'Tracking-Id' in cl:
                    stat = getResult(cl['Tracking-Id'])
                    if 'request_state' in stat and stat['request_state'] == 'SUCCESS':
                        return stat
                    else:
                        for c in stat:
                            l.logDebug(c + ':   ' + str(stat[c]))
                        l.logError('Unsuccesfull request state received.')
                        return None
                else:
                    l.logError('Missing Tracking-Id in response from Joker while updating DS-RR for %s' % zone)
                    return None
                l.logError('Request update DS-RR of zone %s to Joker failed, because: %s' % (zone, stat['result_msg']))
                l.logError('Order Number is %s, and request, sent to Joker was "%s"' % (stat['order number'], q))
                printStatus(stat, ('ds-', 'result_'))
                return None
            else:
                q = str('domain-modify?domain=%s&dnssec=1' % zone)
                i = 1
                for arg in args:
                    if 'pubkey' in arg:
	                    if (None, '') in (arg['alg'], arg['flags'], arg['pubkey']):
	                        l.logError('Internal inconsitency: regAddDS(): at least one argument of key %d is empty: "%d","%d","%d","%s"'
	                            % (arg['tag'], arg['alg'], arg['flags'], urllib.parse.quote(arg['pubkey'])))
	                        return None
	                    q = q + str('&ds-%d=3:%d:%d:%s' %
	                        (i, arg['alg'], arg['flags'], urllib.parse.quote(arg['pubkey'])))
	                    l.logDebug('Unquoted pubkey: %s' % (arg['pubkey'], ))
	                    i = i + 1
                cl = requestJoker(q)
                if not cl: return None
                for c in cl:
                    l.logDebug(c + ':   ' + str(cl[c]))
                if 'Tracking-Id' in cl:
                    stat = getResult(cl['Tracking-Id'])
                    if 'request_state' in stat and stat['request_state'] == 'SUCCESS':
                        return stat
                    else:
                        for c in stat:
                            l.logDebug(c + ':   ' + str(stat[c]))
                        l.logError('Unsuccesfull request state received.')
                        return None
                else:
                    l.logError('Missing Tracking-Id in response from Joker while updating DS-RR for %s' % zone)
                    return None
                l.logError('Request update DS-RR of zone %s to Joker failed, because: %s' % (zone, stat['result_msg']))
                l.logError('Order Number is %s, and request, sent to Joker was "%s"' % (stat['order number'], q))
                printStatus(stat, ('ds-', 'result_'))
                return None
        except (KeyError):
            l.logError('Request update DS-RR of zone %s to Joker failed' % (zone))
            return None
    else:
        l.logError('Internal inconsitency: regAddDS(): submission of DS-RR of .arpa. not implemented')
        return None

def getResultList(rid):
    cl = ''
    if rid == None:
        cl = requestJoker('result-list')
    else:
        cl = requestJoker('result-retrieve?SvTrID=%s' % rid)
    if not cl: return False
    return cl

def getResult(transactionID):
    stat = requestJoker('result-retrieve?SvTrID=%s' % transactionID)
    while stat['Completion-Status'] == '?':
        l.logVerbose('Retrieving completion status from Joker...')
        time.sleep(5)
        stat = requestJoker('result-retrieve?SvTrID=%s' % transactionID)
    stat['TID'] = transactionID
    return stat        

def deleteResult(transactionID):
    stat = requestJoker('result-delete?SvTrID=%s' % transactionID)
    if stat['Status-Code'] != '0':
        l.logVerbose('Deleting completion entry at Joker failed for {}'.format(
                                                                transactionID))
        return None
    return stat        

def printStatus(stat, show):
    lines = []
    for k in stat:
        for s in show:
            if s in k:
                lines.append(str('%s:\t%s\n' % (k, status[k])))
                break
    l.logError(' DNSsec related status info follows: ------------\n%s? End of DNSsec related status info   ------------\n' % ''.join(lines))




"""
 DSKM DNSsec Key Management
 
 Copyright (c) 2012 Axel Rau, axel.rau@chaos1.de

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

# -----------------------------------------
zone.py - the managedZone class module
"""

import sys

from script import path, shell, opts
import script
import fnmatch
from datetime import date, datetime
import time

# for salt
from Crypto import Random as rand
import binascii

import copy

import dns.resolver, dns.message, dns.query, dns.rdatatype, dns.rdtypes.ANY.DNSKEY, dns.rcode
import dns.dnssec, dns.zone

import json
import os

import re
import socket


# -----------------------------------------

import DSKM.registrar as reg
import DSKM.key as dnsKey

import DSKM.logger as logger
l = logger.Logger()

# -----------------------------------------

import DSKM.misc as misc

# -----------------------------------------
# Configurables
# -----------------------------------------
import DSKM.config as conf
#------------------------------------------------------------------------------

ext_recursive_resolver = dns.resolver.Resolver()
ext_recursive_resolver.lifetime = conf.NS_TIMEOUT
ext_recursive_resolver.nameservers = conf.external_recursives
ext_recursive_resolver.use_edns(edns=0, ednsflags=0, payload=4096)

#--------------------------
#   classes
#--------------------------
    
#------------------------------------------------------------------------------
# class managedZone
#------------------------------------------------------------------------------
class managedZone(object):
    """managedZone"""

    def __init__(self, name):
        self.name = name
        
        self.icfg = {'Method': 'unsigned',  # unsigned, NSEC or NSEC3       \
                    'Registrar': 'Local',   # Local, Joker, Ripe            \
                    'Timing': {'ksk': { 'pa': conf.KSK_P_A_INTERVAL,        \
                                        'ai': conf.KSK_A_I_INTERVAL,        \
                                        'id': conf.KSK_I_D_INTERVAL,        \
                                        'i1a2': conf.KSK_I1_A2_INTERVAL     \
                                        },                                  \
                               'zsk': { 'pa': conf.ZSK_P_A_INTERVAL,        \
                                        'ai': conf.ZSK_A_I_INTERVAL,        \
                                        'id': conf.ZSK_I_D_INTERVAL,        \
                                        'i1a2': conf.ZSK_I1_A2_INTERVAL     \
                                        }                                   \
                                }                                           \
                    }
        self.istat = {}
        self.istat['ksk'] = {'State': -1,   # index into state table KSTT, \
                             'Retries': 0}  # Number of retries in current state
        self.istat['zsk'] = {'State': -1,   # index into state table ZSTT, \
                            'Retries': 0}   # Number of retries in current state
        self.istat['OldMethod'] = 'unsigned'# NSEC or NSEC3 \
        self.istat['OldRegistrar'] = 'Local'    # Joker, Ripe

        self.istat['submitted_to_parent'] = []  # list of KSK tags, whose DS have been submitted to parent

        self.pcfg = copy.deepcopy(self.icfg)               # initialize
        self.pstat = copy.deepcopy(self.istat)
        
        self.ksks = []
        self.zsks = []
    
        self.parent = None
        self.parent_dir = None
        
        self.mypath = path(conf.ROOT_PATH + '/' + name)
        self.mypath.cd()
        
        self.remoteDSchanged = False
        self.keys_just_created = []
        
        self.keys_toBeDeleted = []
        
        self.master_DNSKEY_cache = []
        
        #-----------------------------
        # functions in managedZone.__init__
        #-----------------------------
        def readConfig(cfg, domain_name, file_name):
            self.mypath.cd()                    # change to zone directory
            l.logDebug('Opening ' + domain_name + '/' + file_name)
            timingAdded = False
            try:
                with open(file_name) as fd:     # open config/status file for read
                    try:
                        k = ''
                        tstcfg = json.load(fd)
                        for k in iter(cfg):   # do simple syntax check
                            if k == 'Timing' and k not in tstcfg:
                                timingAdded = True
                                tstcfg['Timing'] = copy.deepcopy(cfg['Timing'])
                            vt = tstcfg[k]    # raises if key missed
                            vo = cfg[k]
                            if isinstance(vo, dict):
                                for k in iter(vo):
                                    vo1 = vo[k] # raises if key missed
                                    if isinstance(vo1, dict):
                                        for k in iter(vo1):
                                            vo2 = vo1[k] # raises if key missed
                    except:                     # missing key: syntax error in cfg file
                        l.logError('Garbage found/Missing option ' + k + ' in configuration/status file "' + domain_name + '/' + file_name + '"')
                        e = misc.misc.AbortedZone("")
                        raise e
                    cfg = tstcfg
            except IOError:                     # file not found
                try:
                    with open(file_name, 'w') as fd:
                        json.dump(cfg, fd, indent=8)
                except IOError:                 # no write permission
                    (exc_type, exc_value, exc_traceback) = sys.exc_info()
                    l.logError("Can't create file, because %s" % (exc_value))
                    e = misc.misc.AbortedZone("")
                    raise e
            l.logDebug('Config/status ' + domain_name + '/' + file_name + ' contains:\n' + str(cfg))
            return (cfg, timingAdded)
                
        def deleteKeyFiles():
            self.mypath.cd()                    # change to zone directory
            for kf in os.listdir():             # loop once per file in zone dir
                if fnmatch.fnmatch(kf, 'K' + self.name + '.+*.*'):
                    try:
                        os.remove(path(kf))     # remove all key files if we did not complete zone creation
                        l.logVerbose('Deleted: %s' % (kf))
                    except:
                        (exc_type, exc_value, exc_traceback) = sys.exc_info()
                        l.logError("Can't delete keyfile, because %s" % (exc_value))
                        e = misc.misc.AbortedZone("")
                        raise e
    
        def saveState():
            self.mypath.cd()                    # change to zone directory
            l.logDebug('New status of ' + self.name + ' contains:\n' + str(self.pstat))
            stat_file_name = 'dnssec-stat-' + self.name
            try:
                with open(stat_file_name, 'w') as fd:
                    json.dump(self.pstat, fd, indent=8)
            except:                  # no write permission
            ##except IOError:                 # no write permission
                (exc_type, exc_value, exc_traceback) = sys.exc_info()
                l.logError("Can't create status file, because %s" % (exc_value))
                e = misc.AbortedZone()
                raise e
            
        #-----------------------------
        # end of functions in managedZone.__init__
        #-----------------------------
        (x,y,self.parent) = self.name.partition('.')
        pd = path(conf.ROOT_PATH + '/' + self.parent)
        l.logDebug('Parent directory would be %s' % (pd,))
        zl = ''
        if pd.exists:
            zl = ' <local>'
            self.parent_dir = pd
        if l.verbose:
            print('')
        l.logVerbose('Working at %s on %s (%s %s)' % 
                (datetime.fromtimestamp(time.time()).isoformat(), self.name, self.parent, zl))

        try:
            cfg_file_name = 'dnssec-conf-' + self.name
            (self.pcfg, timingAdded) = readConfig(self.pcfg, name, cfg_file_name)
            if self.pcfg['Timing']['ksk']['ai'] <= self.pcfg['Timing']['ksk']['pa'] + \
                self.pcfg['Timing']['ksk']['i1a2'] + self.pcfg['Timing']['ksk']['id']:
                l.logError('Configuration error in %s: Timimg:ksk:ai must be greater than pa + i1a2 + id' % (cfg_file_name))
                e = misc.AbortedZone("")
                raise e
            if self.pcfg['Timing']['zsk']['ai'] <= self.pcfg['Timing']['zsk']['pa'] + \
                self.pcfg['Timing']['zsk']['i1a2'] + self.pcfg['Timing']['zsk']['id']:
                l.logError('Configuration error in %s: Timimg:zsk:ai must be greater than pa + i1a2 + id' % (cfg_file_name))
                e = misc.AbortedZone("")
                raise e
            if timingAdded:
                self.saveCfgOrState('config')

            stat_file_name = 'dnssec-stat-' + self.name
            (self.pstat, timingAdded) = readConfig(self.pstat, name, stat_file_name)
            if self.pcfg['Method'] not in ('unsigned', 'NSEC', 'NSEC3'):
                l.logError(' Wrong Method "%s" in zone config of %s' % (self.pcfg['Method'], self.name))
                e = misc.AbortedZone("")
                raise e
            if self.pcfg['Registrar'] not in ('Local', 'by hand', 'Joker', 'Ripe'):
                l.logError(' Wrong Registrar "%s" in zone config of %s' % (self.pcfg['Registrar'], self.name))
                e = misc.AbortedZone("")
                raise e
            if self.pstat['OldMethod'] not in ('unsigned', 'NSEC', 'NSEC3'):
                l.logError(' Wrong OldMethod "%s" in zone config of %s' % (self.pstat['Method'], self.name))
                e = misc.AbortedZone("")
                raise e
            
            l.logDebug('KSK state is %d' % (self.pstat['ksk']['State']))
            
            if self.pstat['ksk']['State'] == -1:    # state idle
                deleteKeyFiles()               # delete any key files
            else:
                if self.pstat['OldMethod'] != self.pcfg['Method']:  # don't allow config change if not state idle for now
                    l.logError('Method changed from %s to %s in zone %s' % (self.pstat['OldMethod'], self.pcfg['Method'], self.name))
                    e = misc.AbortedZone('')
                    raise e
                if self.pstat['OldRegistrar'] != self.pcfg['Registrar']:  # don't allow config change if not state idle for now
                    l.logError('Registrar changed from %s to %s in zone %s' % (self.pstat['OldRegistrar'], self.pcfg['Registrar'], self.name))
                    e = misc.AbortedZone('')
                    raise e
                for kf in path('.').list('*'):      # loop once per public key file in zone dir
                    if fnmatch.fnmatch(kf, 'K' + self.name + '.+*.key'):
                        k = dnsKey.SigningKey('read', self.name, kf, self) # and create instance from it
                        if k.type == 'KSK':
                            self.ksks.append(k)
                        elif k.type == 'ZSK':
                            self.zsks.append(k)
            
            nsec3 = False
            if self.pstat['ksk']['State'] == -1: # IDLE state
                if self.pcfg['Method'] == 'unsigned':
                    raise misc.CompletedZone()       # unsigned zone
                elif self.pcfg['Method'] == 'NSEC3':
                    nsec3 = True
                                                # Begin signing zone 1st time
                k = (dnsKey.SigningKey('KSK', self.name, '', self, nsec3=nsec3))
                self.ksks.append(k)
                
                self.zsks.append(dnsKey.SigningKey('ZSK', self.name, '', self, nsec3=nsec3))
            
            for key in self.ksks:
                l.logDebug(key.__str__())
            for key in self.zsks:
                l.logDebug(key.__str__())

            self.pstat['OldMethod'] = self.pcfg['Method']
            self.pstat['OldRegistrar'] = self.pcfg['Registrar']
            
        except misc.AbortedZone:
            l.logError('Aborting zone ' + self.name)
            """
            if self.pstat['ksk']['State'] == -1 or self.pstat['zsk']['State'] == -1:
                try:                # we had an error during read of state/conf
                    saveState()
                except:
                    pass
                l.logError('Removing key files of %s' % (name))
                deleteKeyFiles
                raise
            """
            raise
        pass
    
    #       if self.pcfg['Method'] != self.pstat['OldMethod']:
    #           print('[Method of domain %s has changed from %s to %s]' % (self.name, self.pstat['OldMethod'], self.pcfg['Method']))
    #           script.exit(1, '?Changing of methods not yet implemented')
    #------------------------------------------------------------------------------
    # end of ManagedZone.__init__
    #------------------------------------------------------------------------------
        
    def performStateTransition(self):
        self.keys_just_created = []
        self.remoteDSchanged = False
        self.keys_toBeDeleted = []
        
        zone_loaded = True
        try:
            if misc.doQuery(self.name, 'SOA') == None:
                zone_loaded = False
        except Exception:
            zone_loaded = False
        if not zone_loaded:
            l.logError('Master server does not serve {}.'.format(self.name))
            l.logError('Zone not loaded? Skipping zone.')
            return
        try:
            self.ksks.sort(key=dnsKey.SigningKey.activeTime)
            second = False
            for k in self.ksks:
                if k.state_transition(second):
                    break
                second = True
            
            self.zsks.sort(key=dnsKey.SigningKey.activeTime)
            second = False
            for k in self.zsks:
                if k.state_transition(second):
                    break
                second = True
            
            if self.remoteDSchanged:
                l.logVerbose('About to call registrar. List of keys to request DS-RR: %s ' % (repr(self.pstat['submitted_to_parent'])))
                
                args = self.argsForDSsubmission()
                if len(args) == 0: # removed all DS from remote parents
                    res = reg.regRemoveAllDS(self)
                    if not res:
                        l.logError("Failed to delete all DS-RR of %s at registrar %s" % (self.name, self.pcfg['Registrar']))
                        e = misc.AbortedZone()
                        raise e
                    l.logVerbose("DS-RRs of %s at registrar %s deleted" % (
                            self.name, self.pcfg['Registrar']))
                    for c in res.keys():
                        if c in ('Proc-ID', 'Tracking-Id'):
                            print(c + ':   ' + res[c])
                else:
                    res = reg.regAddDS(self, args)
                    if not res:
                        l.logError("Failed to update DS-RRs for keys %s of %s at registrar %s" % (
                            repr(self.pstat['submitted_to_parent']), self.name, self.pcfg['Registrar']))
                        e = misc.AbortedZone('')
                        raise e
                    l.logVerbose("DS-RRs for keys %s of %s at registrar %s updated" % (
                            repr(self.pstat['submitted_to_parent']), self.name, self.pcfg['Registrar']))
                    for c in res.keys():
                        if c in ('Your mail was received at ......', 'TID'):
                            print(c + ':   ' + res[c])
    
        except misc.AbortedZone:
            l.logError('Aborting zone ' + self.name)
            try:
                for k in self.keys_just_created:
                    self.deleteKeys(k.keytag)
            except:
                raise
                l.logError('Error while deleting recent key files while aborting zone ' + self.name)
        else:                               # don't save state or delete keys if aborted by exception
            for kt in self.keys_toBeDeleted:
                self.deleteKeys(kt)
            self.saveCfgOrState('state')

    
    def argsForDSsubmission(self):
        args = []
        if len(self.pstat['submitted_to_parent']) == 0:
            return args
        for tag in self.pstat['submitted_to_parent']:
            for key in self.ksks:
                if key.keytag == tag:
                    if conf.DIGEST_ALGO_DS == '' or conf.DIGEST_ALGO_DS == '-1':
                        arg = {}
                        arg['tag'] = tag
                        arg['alg'] = key.dnssec_alg
                        arg['digest_type'] = 1
                        arg['digest'] = key.dsHash[0]
                        if conf.DIGEST_ALGO_DS == '-1':
                            arg['flags'] = key.dnssec_flags
                            arg['pubkey'] = key.pubkey_base64
                        args.append(arg)
                    
                    if conf.DIGEST_ALGO_DS == '' or conf.DIGEST_ALGO_DS == '-2':
                        arg = {}
                        arg['tag'] = tag
                        arg['alg'] = key.dnssec_alg
                        arg['digest_type'] = 2
                        arg['digest'] = key.dsHash[1]
                        arg['flags'] = key.dnssec_flags
                        arg['pubkey'] = key.pubkey_base64
                        args.append(arg)
                    
                    break
        return args        
    
    def createFollowUpKey(self, sender):    # usually called by action routine to create a new key
        nsec3 = False
        if self.pcfg['Method'] == 'NSEC3':
            nsec3 = True
        k = (dnsKey.SigningKey(sender.type, self.name, sender.file_name, self, nsec3=nsec3, cloneFromKeyInactiveAt=sender.inactiveTime()))
        if k.type == 'KSK':
             self.ksks.append(k)
        elif k.type == 'ZSK':
             self.zsks.append(k)
        self.keys_just_created.append(k)
        return True
    
    
    def markForDeletion(self, key_tag):
        self.keys_toBeDeleted.append(key_tag)
        return True

    def deleteKeys(self, key_tag):
        l.logVerbose('Deleting one/more of %s' % str(self.mypath.list('K*')))
        self.mypath.cd()              # change to zone directory
        for kf in os.listdir():       # loop once per key file in zone dir
            if key_tag == 0 and fnmatch.fnmatch(kf, 'K' + self.name + '.+*.*') or \
                fnmatch.fnmatch(kf, 'K*' + str(key_tag) + '.*'): # delete all keyfiles or our keyfile
                ##l.logDebug('Matched for deleting: %s' % (kf))
                l.logVerbose('Matched for deleting: %s' % (kf))
                try:
                    os.remove(path(kf))
                    l.logVerbose('Deleted: %s' % (kf))
                except:
                    (exc_type, exc_value, exc_traceback) = sys.exc_info()
                    l.logError("Can't delete keyfile, because %s" % (exc_value))
        if key_tag == 0:
            self.pcfg = copy.deepcopy(self.icfg)               # initialize
            self.pstat = copy.deepcopy(self.istat)
            self.saveCfgOrState('config')
        return True
    
    def UpdateRemoteDS(self, activity, keytag):
        l.logDebug('UpdateRemoteDS called. submitted_to_parent contains: %s ' % (repr(self.pstat['submitted_to_parent'])))
        if activity == 'retire':
            for k in self.pstat['submitted_to_parent']:
                if k != keytag:
                    self.pstat['submitted_to_parent'].remove(k)
        elif activity == 'delete':
            self.pstat['submitted_to_parent'] = []
        elif 'publish' in activity:
            self.pstat['submitted_to_parent'].append(keytag)
        else:
            raise AssertionError('?Wrong activity "%s" in UpdateRemoteDS() with zone %s' % (activity, self.name))
        self.remoteDSchanged = True
        l.logDebug('submitted_to_parent contains now: %s ' % (repr(self.pstat['submitted_to_parent'])))
        return True   
    
    def saveCfgOrState(self, action):       # action is 'config' or 'state'
        self.mypath.cd()                    # change to zone directory
        if action == 'config':
            cfg = self.pcfg
            filename = 'dnssec-conf-' + self.name
        elif action == 'state':
            cfg = self.pstat
            filename = 'dnssec-stat-' + self.name
        else:
            raise AssertionError('?Wrong action "%s" in saveCfgOrState() with zone %s' % (action, self.name))
        
        l.logDebug('Saving config/state in file %s/%s\nNew %s of %s contains:\n %s ' %
                    (os.getcwd(), filename, action, self.name, str(cfg)))
        try:
            with open(filename, 'w') as fd:
                json.dump(cfg, fd, indent=8)
        except:                  # no write permission
        ##except IOError:                 # no write permission
            (exc_type, exc_value, exc_traceback) = sys.exc_info()
            l.logError("Can't create status file, because %s" % (exc_value))
            e = misc.AbortedZone('')
            raise e
    
    def validate(self):                     # validate zone
        global ext_recursive_resolver
         
        if self.pstat['ksk']['State'] < 3 or self.pstat['ksk']['State'] > dnsKey.SigningKey.ksk_state_max or self.pcfg['Registrar'] == 'Local':
            return True
        
        nsAliveTest(self.name)
        
        l.logVerbose('Validating %s...' % (self.name))
        r = ext_recursive_resolver
        try:
            qname = dns.name.from_text(self.name)
            request = dns.message.make_query(qname, rdtype=dns.rdatatype.SOA, rdclass=dns.rdataclass.ANY)
            request.use_edns(r.edns, r.ednsflags, r.payload)
            request.want_dnssec(True)
            response = None
            nameservers = r.nameservers[:]
            for nameserver in nameservers[:]:
                try:
                    l.logDebug('Querying %s via UDP' % (nameserver, ))
                    response = dns.query.udp(request, nameserver, conf.NS_TIMEOUT)
                    if response.flags & dns.flags.TC:
                        # Response truncated; retry with TCP.
                        l.logDebug('Querying %s via TCP' % (nameserver, ))
                        response = dns.query.tcp(request, nameserver, conf.NS_TIMEOUT)
                except (socket.error, dns.exception.Timeout, dns.query.UnexpectedSource, dns.exception.FormError, EOFError):
                    l.logDebug('Queryerror')
                    response = None
                    continue
                rcode = response.rcode()
                if rcode == dns.rcode.NOERROR or \
                       rcode == dns.rcode.NXDOMAIN:
                    l.logDebug('Query gave NOERROR or NXDOMAIN')
                    break
                response = None
            ## end of for nameserver in r.nameservers

            if response is None:
                e = dns.resolver.NoAnswer
                raise e
            ans = dns.resolver.Answer(qname, dns.rdatatype.SOA, dns.rdataclass.ANY, response, False)

            l.logDebug('Flags are %x; %x should be set' % (ans.response.flags, dns.flags.AD))
            if ans.response.flags & dns.flags.AD:
                l.logVerbose('OK')
                return True
        except (dns.resolver.NoAnswer, dns.resolver.NXDOMAIN):
            pass
        except (dns.exception.Timeout):
            (exc_type, exc_value, exc_traceback) = sys.exc_info()
            errmsg = "%s: DS query timed out. %s, %s" % \
                (self.name, exc_type, exc_value)
            l.logError('Validation of %s FAILED' % (self.name))
            l.logError(errmsg)
        l.logError('Validation of %s FAILED' % (self.name))
        return False
    
    # compare SOA serial of master with zone file (bind 9.11 may be out of sync) 
    def verifySerial(self):
        try:
            answer = misc.doQuery(self.name, 'SOA')
        except Exception:
            l.logError('Failed to compare serial with zone file')
            return False
        result = [ str(rdata) for rdata in answer ][0].split()[2]
        
        filename = self.mypath + '/' + self.name + '.zone'
        with open(filename, 'r', encoding="ASCII") as fd:
            try:
                zf = fd.read()
            except:                 # file not found or not readable
                raise Exception("verifySerial can't read zone file " + filename)
        sea = re.search('(\d{10})(\s*;\s*)(Serial number)', zf)
        serial = sea.group(1)
                
        l.logDebug('verifySerial of {}: file: {} dns: {}'.format(self.name, serial, result))
        if int(serial) > int(result):   # file more recent than master server?
            l.logError('Serial of {} out of sync: {} > {}'.format(self.name, serial, result))
            return False
        return True
    
    def stopSigning(self, force):
        ##import pdb;pdb.set_trace()
        if not force:
            if self.pstat['ksk']['State'] < 2:
                l.logError("Can't stop signing in state %s" % (self.pstat['ksk']['State']))
                return 1
            if self.pstat['ksk']['State'] > dnsKey.SigningKey.ksk_state_max:
                print("%%Termination of signing already in progress (state=%s)" % (self.pstat['ksk']['State']))
                return 0
        secondKey = False
        for k in self.ksks:
            k.UpdateDS('delete', False) # remove all DS-RR in parent zone
            l.logVerbose("DS-RRs of %s at registrar %s deleted" % (
                            self.name, self.pcfg['Registrar']))
            k.set_delete_time(secondKey)
            secondKey = True
        secondKey = False
        for k in self.zsks:
            k.set_delete_time(secondKey)
            secondKey = True
        self.pstat['ksk']['State'] = dnsKey.SigningKey.ksk_state_max + 1
        if force:
            self.deleteKeys(0)
            try:
                res = str(shell(str('rndc loadkeys %s' % (self.name)), stderr='PIPE').stderr)
                l.logDebug('Rndc loadkeys returned: %s' % (res))
            except script.CommandFailed:
                l.logError('Error during rndc loadkeys after deleting keys of %s' % (self.name))
        self.pstat['ksk']['Retries'] = 0
        self.pstat['zsk']['Retries'] = 0
        self.saveCfgOrState('state')
        if self.remoteDSchanged and len(self.pstat['submitted_to_parent']) > 0:
            l.logVerbose('About to call registrar. List of keys to request DS-RR: %s ' % (repr(self.pstat['submitted_to_parent'])))
            res = reg.regRemoveAllDS(self)
            if not res:
                em = str("Failed to delete all DS-RR of %s at registrar %s" % (self.name, self.pcfg['Registrar']))
                l.logError(em)
                e = misc.AbortedZone(em)
                raise e
            l.logVerbose("DS-RRs of %s at registrar %s deleted" % (
                    self.name, self.pcfg['Registrar']))
            for c in res.keys():
                if c in ('Proc-ID', 'Tracking-Id'):
                    print(c + ':   ' + res[c])
        return 0         


def nsAliveTest(theZone):         # query all authoritative NS for SOA of zone
    global ext_recursive_resolver
        
    r = ext_recursive_resolver
    
    qname = dns.name.from_text(theZone)
    request = dns.message.make_query(qname, rdtype=dns.rdatatype.SOA, rdclass=dns.rdataclass.ANY)
    request.use_edns(r.edns, r.ednsflags, r.payload)
    request.want_dnssec(True)
    response = None
    nameservers = misc.authNS(theZone)
    for nameserver in nameservers[:]:
        try:
            l.logDebug('Querying {} for SOA of {} via TCP'.format(nameserver, theZone))
            response = dns.query.tcp(request, nameserver, 10)
            rcode = response.rcode()
            if rcode == 0: continue
        except (socket.error, dns.exception.Timeout, dns.query.UnexpectedSource,
                dns.exception.FormError, EOFError, dns.resolver.NoAnswer,
                dns.resolver.NXDOMAIN):
            pass
        l.logError('NS {} did not answer SOA query for {}'.format(
                                                    nameserver, theZone))

"""
r = misc.authResolver('2.0.0.0.0.4.d.0.2.0.a.2.ip6.arpa')
for ns in r.nameservers:
    print(ns)
"""


"""
 DSKM DNSsec Key Management
 
 Copyright (c) 2012 Axel Rau, axel.rau@chaos1.de

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

# -----------------------------------------
registrar.py - common registrar functions
"""

# -----------------------------------------
import http.client
import time
import urllib.parse

import sys
# -----------------------------------------

# -----------------------------------------
# Globals
# -----------------------------------------
theConnection = None
# -----------------------------------------

import DSKM.logger as logger
l = logger.Logger()
# -----------------------------------------


# -----------------------------------------
# Configurables
# -----------------------------------------
import DSKM.config as conf
import DSKM.REG.joker as reg_joker
import DSKM.REG.ripe as reg_ripe

#------------------------------------------------------------------------------

# -----------------------------------------
# Functions
# -----------------------------------------
def regRemoveAllDS(zone):
    zone_name = zone.name
    if zone.pcfg['Registrar'] == 'Joker':
        return reg_joker.regRemoveAllDS(zone_name)
    elif zone.pcfg['Registrar'] == 'Ripe':
        return reg_ripe.regRemoveAllDS(zone_name)
    elif zone.pcfg['Registrar'] == 'by hand':
        return handOverByEmail(zone_name, [], str('Deletion of DS-RR of zone %s required' % zone_name))
    else:
        l.logError('Internal inconsistency: Unknown registrar "%s" in config' % (zone.pcfg['Registrar']))
    return None
    
def regAddDS(zone, args):
    zone_name = zone.name
    if zone.pcfg['Registrar'] == 'Joker':
        return reg_joker.regAddDS(zone_name, args)
    elif zone.pcfg['Registrar'] == 'Ripe':
        return reg_ripe.regAddDS(zone_name, args, False, False)
    elif zone.pcfg['Registrar'] == 'by hand':
        return handOverByEmail(zone_name, args, str('DS-RR handover to parent of zone %s required' % zone_name))
    else:
        l.logError('Internal inconsistency: Unknown registrar "%s" in config' % (zone.pcfg['Registrar']))
    return None

def getResultList(rid):
    return reg_joker.getResultList(rid)
    l.logWarn('No resultlists at Ripe')

def deleteResult(transactionID):
    return reg_joker.deleteResult(transactionID)
    l.logWarn("Can't delete result info at Ripe")

def regTest(zone, dry_run):
    zone_name = zone.name
    if zone.pcfg['Registrar'] == 'Joker':
        ##return reg_joker.regTest(zone_name, dry_run)
        return
    elif zone.pcfg['Registrar'] == 'Ripe':
        args = zone.argsForDSsubmission()
        return reg_ripe.regAddDS(zone_name, args, True, dry_run)
    else:
        return

def handOverByEmail(zone_name, args, subject):
    body = ''
    if len(args) == 0:
        body = str('Please ask parent zone operator to remove all DS-RR for zone\n\t%s\n from parent zone.\n' % (zone_name))
    else:
        body = str('Please ask parent zone operator to adjust (delete/add) DS-RR / DNSKEY-RR for zone\n\t%s\n.\n' % (zone_name))
        body = body + '''Inspect the following list of DS-RR / DNSKEY-RR pairs carefully.
Any pairs, existing at parent and not in the following list must be deleted.
Any pairs, missing at parent, must be added.\n'''
        i = 1
        for arg in args:
            if (None, '') in (arg['tag'], arg['alg'], arg['digest_type'], str(arg['digest'])):
                l.logError('Internal inconsitency: regAddDS(): at least one argument of key %d is empty: "%d","%d","%s"'
                    % (arg['tag'], arg['alg'], arg['digest_type'], str(arg['digest'])))
                return None
            body = body + str('\n\nDS-RR / DNSKEY-RR %d -------------------------------------\n' % (i))
            body = body + str('DS\t%s %s %s\n' % (arg['tag'], arg['digest_type'], arg['digest']))
            body = body + str('DNSKEY\t%s 3 8 %s\n' % (arg['flags'], arg['pubkey']))

            i = i + 1
    body = body + str('\nEnd of message ----------------------------------\n')
    l.sendMail(subject, body)
    return {'TID': 'E-Mail sent'}



"""
Copyright (C) 2015-2017  Axel Rau <axel.rau@chaos1.de>

This file is part of serverPKI.

serverPKI is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Foobar is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with serverPKI.  If not, see <http://www.gnu.org/licenses/>.
"""


# Module to make config settings available to members of package

import sys

# name of our config module
CONFIG_MODULE = 'config'

# place where to find it (sys.prefix point at venve if we are in a venv)
CONFIG_MODULE_DIRS =(   sys.prefix + '/etc',
                        '/usr/local/etc/DSKM')

sys.path.append(CONFIG_MODULE_DIRS[0])
sys.path.append(CONFIG_MODULE_DIRS[1])

from dskm_conf import *



"""
 DSKM DNSsec Key Management
 
 Copyright (c) 2012 Axel Rau, axel.rau@chaos1.de

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

# -----------------------------------------
misc.py - miscellaneous classes and functions
"""

import dns.resolver, dns.message, dns.query, dns.rdatatype, dns.rdtypes.ANY.DNSKEY, dns.rcode
import dns.dnssec, dns.zone

import sys

import DSKM.config as conf
import DSKM.key

import DSKM.logger as logger
l = logger.Logger()


#--------------------------

auth_NS = {}
auth_resolver = {}

# names and instances of zones, setup by main
zone_dirs = []
zones = {}

#--------------------------
#   classes
#--------------------------
# exceptions

class AbortedZone(Exception):
    def __init__(self,x):
        self.data = x

class CompletedZone(Exception):
    pass
    


#--------------------------
#   functions
#--------------------------

def doQuery(theQuery, theRRtype):
    try:
        answer = DSKM.key.master_resolver.query(theQuery, theRRtype)
        return answer
    except (dns.resolver.NoAnswer, dns.resolver.NXDOMAIN, KeyError):
        (exc_type, exc_value, exc_traceback) = sys.exc_info()
        l.logDebug('doQuery(): Query: %s failed with %s' % (theQuery, exc_type))
        return None
    except (dns.exception.Timeout):
        l.logError('Failed to query %s of zone %s (Timeout)' % (theRRtype, repr(theQuery)))
        raise
    except (dns.resolver.NoNameservers):
        l.logError('Failed to query %s of zone %s (NoNameservers)' % (theRRtype, repr(theQuery)))
        raise

def authNS(theZone):        # return list of NS addresses, authoritative for theZone
    global auth_NS
    
    if theZone in auth_NS:
        return auth_NS[theZone]
    
    nslist = []
    ll = theZone.split('.')
    ll.append('')
    a1 = None
    while len(ll) > 1:
    	n = dns.name.Name(ll)
    	a1 = doQuery(n, 'NS')
    	if a1:
    		break
    	del(ll[0])
    if a1:
        for ns in a1:
            a2 = doQuery(ns.target, 'A')
            if a2:
                nslist.append(a2[0].address)
            a2 = doQuery(ns.target, 'AAAA')
            if a2:
                nslist.append(a2[0].address)
        auth_NS[theZone] = nslist
        return nslist
    l.logWarn("Unable to find NS of zone %s (or it's parent" % (repr(theQuery)))
    e = AbortedZone("")
    raise e

def authResolver(theZone):        # return a resolver bound to NS addresses, authoritative for theZone
    global auth_resolver
    if theZone in auth_resolver:
        return auth_resolver[theZone]
    my_resolver = dns.resolver.Resolver()
    my_resolver.lifetime = conf.NS_TIMEOUT
    my_resolver.nameservers = authNS(theZone)
    my_resolver.use_edns(edns=0, ednsflags=0, payload=4096)
    auth_resolver[theZone] = my_resolver
    return my_resolver



"""
 DSKM DNSsec Key Management
 
 Copyright (c) 2012-2019 Axel Rau, axel.rau@chaos1.de

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

# -----------------------------------------
# utility module of serverPKI (commandline parsing, logging ...)
# -----------------------------------------
"""



#--------------- imported modules --------------
import optparse


#--------- globals ***DO WE NEED THIS?*** ----------

global options


#--------------- command line options --------------

parser = optparse.OptionParser(description='DSKM DNSsec Key Management\n'
                    'Do maintenace of DNSsec keys.\n'
                    'Create and delete them as necessary.\n'
                    'Submit/cancle DS-RR to/at parent registrar.')

parser.add_option('--cron', '-c', dest='cron', action='store_true',
                   default=False,
                   help='Run as cronjob. Each run increments timeout timer.')

parser.add_option('--stopSigningOfZone', '-S', action='store',
                   help=('Initiate procedure to make a zone unsigned. '
                   'Argument is zone name.'))
                   
parser.add_option('--force', '-f', action='store_true',
                   default=False,
                   help='Force deletion of keys (ignore delete time) while stopping signing of zone.'),

parser.add_option('--registrar_status', '-r', action='store_true',
                   default=False,
                   help='Query list of completed and pending requests of all registrars and terminate.'),

parser.add_option('--purge_all_registrar_completion_info', '-p', action='store_true',
                   default=False,
                   help='Purge all completion info of completed and pending requests of all registrars and terminate.'),

parser.add_option('--query_status', '-q', action='store',
                   help='Give detailed registrar result status about <request-id>.'),

parser.add_option('--test_registrar_DS_submission', '-t', action='store_true',
                   default=False,
                   help='Delete and re-submit current DS-RR to registrar.')

parser.add_option('--dry-run', '-n', dest='dry_run', action='store_true',
                   default=False,
                   help=('Do not really change any data at registrar with '
                        '--test_registrar_DS_submission.')),

parser.add_option('--debug', '-d', action='store_true',
                   default=False,
                   help='Turn on debugging.'),
parser.add_option('--verbose', '-v', dest='verbose', action='store_true',
                   default=False,
                   help='Be more verbose.')

options, args = parser.parse_args()

if options.debug: options.verbose = True





import setuptools

# Version info -- read without importing
_locals = {}
with open('DSKM/_version.py') as fp:
    exec(fp.read(), None, _locals)
version = _locals['__version__']

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name = "DSKM",
    version = version,
    author = "Axel Rau",
    author_email = "axel.rau@chaos1.de",
    description = "DNSsec key management",
    long_description = long_description,
    long_description_content_type="text/markdown",
    url = "https://github.com/mc3/DSKM",
    packages = [ 'DSKM' ],
    data_files =[('share/doc/DSKM', ['dnssec_key_states.graffle',
                                          'dnssec_key_states.pdf'])],                                      
    entry_points = {
        'console_scripts': [
            'operate_dskm = DSKM.operate:execute_from_command_line',
        ],
    },
    license = 'GPLv3',
    classifiers = [
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Operating System :: POSIX',
        'Topic :: Internet',
        'Topic :: Security :: Cryptography',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Natural Language :: English',
    ],
    install_requires=[
        'dnspython>=1.16.0',
        'ecdsa>=0.13',  
        'pycryptodome>=3.7.3',
        'script>=1.7.2',
    ],
)



"""
 DSKM DNSsec Key Management
 
 Copyright (c) 2012 Axel Rau, axel.rau@chaos1.de

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

# -----------------------------------------
logger.py - Logger class module - centraliced logging and alarming
"""

import pprint
import smtplib
from email.mime.text import MIMEText

# -----------------------------------------
import DSKM.config as conf

#--------------------------
#   classes
#--------------------------

class Logger():
    """Log"""
    
    _singleton = None
    debug = False
    verbose = False
    cron = False
    debugText = ''
    verboseText = ''
    lastError = ''
    lastWarning = ''
    
    def __new__(cls, *args, **kwargs):
        if not cls._singleton:
            cls._singleton = super(Logger, cls ).__new__(cls, *args, **kwargs)
        return cls._singleton
    
    
    def __init__(self, verbose=True, debug=False, cron=False):
        
        Logger.debug = debug
        Logger.verbose = verbose
        Logger.cron = cron    
    
    def logError(self, text):           # Fatal error
        Logger.lastError = '?%s' % (text)
        print(Logger.lastError)
        Logger.debugText = Logger.debugText + Logger.lastError + '\n'

    
    def logWarn(self, text):            # Warning
        Logger.lastWarning = '%%%s' % (text)
        print(Logger.lastWarning)
        Logger.verboseText = Logger.verboseText + Logger.lastWarning + '\n'
        Logger.debugText = Logger.debugText + Logger.lastWarning + '\n'
    
    def logVerbose(self, text):
        im = '[%s]' % (text)            # informal message
        if Logger.verbose:
            print(im)
        Logger.verboseText = Logger.verboseText + im + '\n'
        Logger.debugText = Logger.debugText + im + '\n'
    
    def logDebug(self, text, level=0):
        dm = '[%s]' % (text)            # debug message
        if Logger.debug:
            print(dm)
        Logger.debugText = Logger.debugText + dm + '\n'
    
    def mailErrors(self):               # called by main on exit
        if len(Logger.lastError) > 0:
            self.sendMail(Logger.lastError, Logger.debugText, True)
        elif len(Logger.lastWarning) > 0:
            self.sendMail(Logger.lastWarning, Logger.verboseText, True)
        

    def sendMail(self, subject, body, onlyCron=False):
        if not conf.mailRelay:           # done, if not configured
            return
        if onlyCron and not Logger.cron: # mail only if cronjob, if so requested
            return
        msg = MIMEText(body)
        msg['Subject'] = '[DSKM] ' + subject
        msg['From'] = conf.sender
        msg['To'] = ', '.join(conf.recipients)
        s = smtplib.SMTP(conf.mailRelay)
        s.send_message(msg)
        s.quit
    
    def cronjob(self):
        return Logger.cron
